import logging
from datetime import date
from typing import Any, List, Optional

import pandas as pd
import pydantic
from arq import ArqRedis
from fastapi import APIRouter, Depends, File, Query, UploadFile, status
from fastapi.encoders import jsonable_encoder
from fastapi.exceptions import HTTPException
from fastapi.responses import StreamingResponse
from sqlalchemy import null, or_, select
from sqlalchemy.exc import MultipleResultsFound, NoResultFound
from sqlalchemy.orm import Session

from mavedb import deps
from mavedb.lib.authentication import UserData
from mavedb.lib.authorization import (
    get_current_user,
    require_current_user,
    require_current_user_with_email,
    RoleRequirer,
)
from mavedb.lib.contributors import find_or_create_contributor
from mavedb.lib.exceptions import MixedTargetError, NonexistentOrcidUserError, ValidationError
from mavedb.lib.identifiers import (
    create_external_gene_identifier_offset,
    find_or_create_doi_identifier,
    find_or_create_publication_identifier,
)
from mavedb.lib.logging import LoggedRoute
from mavedb.lib.logging.context import (
    correlation_id_for_context,
    logging_context,
    save_to_logging_context,
)
from mavedb.lib.permissions import Action, assert_permission, has_permission
from mavedb.lib.score_sets import (
    csv_data_to_df,
    find_meta_analyses_for_experiment_sets,
    get_score_set_counts_as_csv,
    get_score_set_scores_as_csv,
    variants_to_csv_rows,
)
from mavedb.lib.score_sets import (
    fetch_superseding_score_set_in_search_result,
    search_score_sets as _search_score_sets,
    refresh_variant_urns,
)
from mavedb.lib.taxonomies import find_or_create_taxonomy
from mavedb.lib.urns import (
    generate_experiment_set_urn,
    generate_experiment_urn,
    generate_score_set_urn,
)
from mavedb.models.contributor import Contributor
from mavedb.models.enums.processing_state import ProcessingState
from mavedb.models.enums.user_role import UserRole
from mavedb.models.experiment import Experiment
from mavedb.models.license import License
from mavedb.models.mapped_variant import MappedVariant
from mavedb.models.score_set import ScoreSet
from mavedb.models.target_accession import TargetAccession
from mavedb.models.target_gene import TargetGene
from mavedb.models.target_sequence import TargetSequence
from mavedb.models.variant import Variant
from mavedb.view_models import mapped_variant, score_set, calibration
from mavedb.view_models.search import ScoreSetsSearch

logger = logging.getLogger(__name__)


async def fetch_score_set_by_urn(
    db, urn: str, user: Optional[UserData], owner_or_contributor: Optional[UserData], only_published: bool
) -> Optional[ScoreSet]:
    """
    Fetch one score set by URN, ensuring that the user has read permission.

    :param db: An active database session.
    :param urn: The score set URN.
    :param user: The user who has requested the score set. If the user does not have read permission, the score set will
      not be returned. If None, the score set is returned only if publicly visible.
    :param owner_or_contributor: If not None, require that the result be a score set of which this user is owner or
      contributor.
    :param only_published: If true, only return the score set if it is published.
    :return: The score set, or None if the URL was not found or refers to a private score set not owned by the specified
        user.
    """
    try:
        query = db.query(ScoreSet).filter(ScoreSet.urn == urn)
        if owner_or_contributor is not None:
            query.filter(
                or_(
                    ScoreSet.private.is_(False),
                    ScoreSet.created_by_id == owner_or_contributor.user.id,
                    ScoreSet.contributors.any(Contributor.orcid_id == owner_or_contributor.user.username),
                )
            )
        if only_published:
            query.filter(ScoreSet.private.is_(False))
        item = query.one_or_none()
    except MultipleResultsFound:
        logger.info(
            msg="Could not fetch the requested score set; Multiple such score sets exist.", extra=logging_context()
        )
        raise HTTPException(status_code=500, detail=f"multiple score sets with URN '{urn}' were found")

    if not item:
        logger.info(msg="Could not fetch the requested score set; No such score sets exist.", extra=logging_context())
        raise HTTPException(status_code=404, detail=f"score set with URN '{urn}' not found")

    assert_permission(user, item, Action.READ)

    if item.superseding_score_set and not has_permission(user, item.superseding_score_set, Action.READ).permitted:
        item.superseding_score_set = None

    return item


router = APIRouter(
    prefix="/api/v1",
    tags=["score sets"],
    responses={404: {"description": "not found"}},
    route_class=LoggedRoute,
)


@router.post("/score-sets/search", status_code=200, response_model=list[score_set.ShortScoreSet])
def search_score_sets(
    search: ScoreSetsSearch,
    db: Session = Depends(deps.get_db),
    user_data: Optional[UserData] = Depends(get_current_user),
) -> Any:  # = Body(..., embed=True),
    """
    Search score sets.
    """
    score_sets = _search_score_sets(db, None, search)
    return fetch_superseding_score_set_in_search_result(score_sets, user_data, search)


@router.get("/score-sets/mapped-genes", status_code=200, response_model=dict[str, list[str]])
def score_set_mapped_gene_mapping(
    db: Session = Depends(deps.get_db), user_data: UserData = Depends(get_current_user)
) -> Any:
    """
    Get a mapping of score set URNs to mapped gene symbols.
    """
    save_to_logging_context({"requested_resource": "mapped-genes"})

    score_sets_with_mapping_metadata = db.execute(
        select(ScoreSet, TargetGene.post_mapped_metadata)
        .join(ScoreSet)
        .where(TargetGene.post_mapped_metadata.is_not(None))
    ).all()

    mapped_genes: dict[str, list[str]] = {}
    for score_set_item, post_mapped_metadata in score_sets_with_mapping_metadata:
        if not has_permission(user_data, score_set_item, Action.READ).permitted:
            continue

        sequence_genes = [
            *post_mapped_metadata.get("genomic", {}).get("sequence_genes", []),
            *post_mapped_metadata.get("protein", {}).get("sequence_genes", []),
        ]

        if sequence_genes:
            mapped_genes.setdefault(score_set_item.urn, []).extend(sequence_genes)

    return mapped_genes


@router.post(
    "/me/score-sets/search",
    status_code=200,
    response_model=list[score_set.ShortScoreSet],
)
def search_my_score_sets(
    search: ScoreSetsSearch,  # = Body(..., embed=True),
    db: Session = Depends(deps.get_db),
    user_data: UserData = Depends(require_current_user),
) -> Any:
    """
    Search score sets created by the current user..
    """
    score_sets = _search_score_sets(db, user_data.user, search)
    return fetch_superseding_score_set_in_search_result(score_sets, user_data, search)


@router.get(
    "/score-sets/{urn}",
    status_code=200,
    response_model=score_set.ScoreSet,
    responses={404: {}, 500: {}},
    response_model_exclude_none=True,
)
async def show_score_set(
    *,
    urn: str,
    db: Session = Depends(deps.get_db),
    user_data: UserData = Depends(get_current_user),
) -> Any:
    """
    Fetch a single score set by URN.
    """
    save_to_logging_context({"requested_resource": urn})
    return await fetch_score_set_by_urn(db, urn, user_data, None, False)


@router.get(
    "/score-sets/{urn}/scores",
    status_code=200,
    responses={
        200: {
            "content": {"text/csv": {}},
            "description": """Variant scores in CSV format, with four fixed columns (accession, hgvs_nt, hgvs_pro,"""
            """ and hgvs_splice), plus score columns defined by the score set.""",
        }
    },
)
def get_score_set_scores_csv(
    *,
    urn: str,
    start: int = Query(default=None, description="Start index for pagination"),
    limit: int = Query(default=None, description="Number of variants to return"),
    drop_na_columns: Optional[bool] = None,
    db: Session = Depends(deps.get_db),
    user_data: Optional[UserData] = Depends(get_current_user),
) -> Any:
    """
    Return scores from a score set, identified by URN, in CSV format.
    If no start and limit, all of variants of this score set will be returned.
    Example path:
    /score-sets/{urn}/scores
    /score-sets/{urn}/scores?start=0&limit=100
    /score-sets/{urn}/scores?start=100
    """
    save_to_logging_context(
        {
            "requested_resource": urn,
            "resource_property": "scores",
            "start": start,
            "limit": limit,
        }
    )

    if start and start < 0:
        logger.info(msg="Could not fetch scores with negative start index.", extra=logging_context())
        raise HTTPException(status_code=400, detail="Start index must be non-negative")
    if limit is not None and limit <= 0:
        logger.info(msg="Could not fetch scores with non-positive limit.", extra=logging_context())
        raise HTTPException(status_code=400, detail="Limit must be positive")

    score_set = db.query(ScoreSet).filter(ScoreSet.urn == urn).first()
    if not score_set:
        logger.info(msg="Could not fetch the requested scores; No such score set exists.", extra=logging_context())
        raise HTTPException(status_code=404, detail=f"score set with URN '{urn}' not found")

    assert_permission(user_data, score_set, Action.READ)

    csv_str = get_score_set_scores_as_csv(db, score_set, start, limit, drop_na_columns)
    return StreamingResponse(iter([csv_str]), media_type="text/csv")


@router.get(
    "/score-sets/{urn}/counts",
    status_code=200,
    responses={
        200: {
            "content": {"text/csv": {}},
            "description": """Variant counts in CSV format, with four fixed columns (accession, hgvs_nt, hgvs_pro,"""
            """ and hgvs_splice), plus score columns defined by the score set.""",
        }
    },
)
async def get_score_set_counts_csv(
    *,
    urn: str,
    start: int = Query(default=None, description="Start index for pagination"),
    limit: int = Query(default=None, description="Number of variants to return"),
    drop_na_columns: Optional[bool] = None,
    db: Session = Depends(deps.get_db),
    user_data: Optional[UserData] = Depends(get_current_user),
) -> Any:
    """
    Return counts from a score set, identified by URN, in CSV format.
    If no start and limit, all of variants of this score set will be returned.
    Example path:
    /score-sets/{urn}/counts
    /score-sets/{urn}/counts?start=0&limit=100
    /score-sets/{urn}/counts?start=100
    """
    save_to_logging_context(
        {
            "requested_resource": urn,
            "resource_property": "counts",
            "start": start,
            "limit": limit,
        }
    )

    if start and start < 0:
        logger.info(msg="Could not fetch counts with negative start index.", extra=logging_context())
        raise HTTPException(status_code=400, detail="Start index must be non-negative")
    if limit is not None and limit <= 0:
        logger.info(msg="Could not fetch counts with non-positive limit.", extra=logging_context())
        raise HTTPException(status_code=400, detail="Limit must be positive")

    score_set = db.query(ScoreSet).filter(ScoreSet.urn == urn).first()
    if not score_set:
        logger.info(msg="Could not fetch the requested counts; No such score set exist.", extra=logging_context())
        raise HTTPException(status_code=404, detail=f"score set with URN {urn} not found")

    assert_permission(user_data, score_set, Action.READ)

    csv_str = get_score_set_counts_as_csv(db, score_set, start, limit, drop_na_columns)
    return StreamingResponse(iter([csv_str]), media_type="text/csv")


@router.get(
    "/score-sets/{urn}/mapped-variants",
    status_code=200,
    response_model=list[mapped_variant.MappedVariant],
)
def get_score_set_mapped_variants(
    *,
    urn: str,
    db: Session = Depends(deps.get_db),
    user_data: Optional[UserData] = Depends(get_current_user),
) -> Any:
    """
    Return mapped variants from a score set, identified by URN.
    """
    save_to_logging_context({"requested_resource": urn, "resource_property": "mapped-variants"})

    score_set = db.query(ScoreSet).filter(ScoreSet.urn == urn).first()
    if not score_set:
        logger.info(
            msg="Could not fetch the requested mapped variants; No such score set exist.", extra=logging_context()
        )
        raise HTTPException(status_code=404, detail=f"score set with URN {urn} not found")

    assert_permission(user_data, score_set, Action.READ)

    mapped_variants = (
        db.query(MappedVariant)
        .filter(ScoreSet.urn == urn)
        .filter(ScoreSet.id == Variant.score_set_id)
        .filter(Variant.id == MappedVariant.variant_id)
        .all()
    )

    if not mapped_variants:
        logger.info(msg="No mapped variants are associated with the requested score set.", extra=logging_context())
        raise HTTPException(
            status_code=404,
            detail=f"No mapped variant associated with score set URN {urn} was found",
        )

    return mapped_variants


@router.post(
    "/score-sets/",
    response_model=score_set.ScoreSet,
    responses={422: {}},
    response_model_exclude_none=True,
)
async def create_score_set(
    *,
    item_create: score_set.ScoreSetCreate,
    db: Session = Depends(deps.get_db),
    user_data: UserData = Depends(require_current_user_with_email),
) -> Any:
    """
    Create a score set.
    """
    logger.debug(msg="Began score set creation.", extra=logging_context())

    experiment: Optional[Experiment] = None
    if item_create.experiment_urn is not None:
        experiment = db.query(Experiment).filter(Experiment.urn == item_create.experiment_urn).one_or_none()
        if not experiment:
            logger.info(
                msg="Failed to create score set; The requested experiment does not exist.", extra=logging_context()
            )
            raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail="Unknown experiment")
        # Not allow add score set in meta-analysis experiments.
        if any(s.meta_analyzes_score_sets for s in experiment.score_sets):
            raise HTTPException(
                status_code=status.HTTP_403_FORBIDDEN,
                detail="Score sets may not be added to a meta-analysis experiment.",
            )

        save_to_logging_context({"experiment": experiment.urn})
        assert_permission(user_data, experiment, Action.ADD_SCORE_SET)

    license_ = db.query(License).filter(License.id == item_create.license_id).one_or_none()
    save_to_logging_context({"requested_license": item_create.license_id})

    if not license_:
        logger.info(msg="Failed to create score set; The requested license does not exist.", extra=logging_context())
        raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail="Unknown license")
    elif not license_.active:
        logger.info(
            msg="Failed to create score set; The requested license is no longer active.", extra=logging_context()
        )
        raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail="Invalid license")

    save_to_logging_context({"requested_superseded_score_set": item_create.superseded_score_set_urn})
    if item_create.superseded_score_set_urn is not None:
        superseded_score_set = await fetch_score_set_by_urn(
            db, item_create.superseded_score_set_urn, user_data, user_data, True
        )

        if superseded_score_set is None:
            logger.info(
                msg="Failed to create score set; The requested superseded score set does not exist.",
                extra=logging_context(),
            )
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail="Unknown superseded score set",
            )
    else:
        superseded_score_set = None

    distinct_meta_analyzes_score_set_urns = list(set(item_create.meta_analyzes_score_set_urns or []))
    meta_analyzes_score_sets = [
        ss
        for ss in [
            await fetch_score_set_by_urn(db, urn, user_data, None, True)
            for urn in distinct_meta_analyzes_score_set_urns
        ]
        if ss is not None
    ]

    save_to_logging_context({"requested_meta_analyzes_score_sets": distinct_meta_analyzes_score_set_urns})
    for i, meta_analyzes_score_set in enumerate(meta_analyzes_score_sets):
        if meta_analyzes_score_set is None:
            logger.info(
                msg=f"Failed to create score set; The requested meta analyzed score set ({distinct_meta_analyzes_score_set_urns[i]}) does not exist.",
                extra=logging_context(),
            )
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail=f"Unknown meta-analyzed score set {distinct_meta_analyzes_score_set_urns[i]}",
            )

    if len(meta_analyzes_score_sets) > 0:
        # If any existing score set is a meta-analysis for score sets in the same collection of experiment sets, use its
        # experiment as the parent of our new meta-analysis. Otherwise, create a new experiment.
        meta_analyzes_experiment_sets = list(
            set(
                (
                    ss.experiment.experiment_set
                    for ss in meta_analyzes_score_sets
                    if ss.experiment.experiment_set is not None
                )
            )
        )
        meta_analyzes_experiment_set_urns = [es.urn for es in meta_analyzes_experiment_sets if es.urn is not None]
        existing_meta_analyses = find_meta_analyses_for_experiment_sets(db, meta_analyzes_experiment_set_urns)

        if len(existing_meta_analyses) > 0:
            experiment = existing_meta_analyses[0].experiment
        elif len(meta_analyzes_experiment_sets) == 1:
            # The analyzed score sets all belong to one experiment set, so the meta-analysis should go in that
            # experiment set's meta-analysis experiment. But there is no meta-analysis experiment (or else we would
            # have found it by looking at existing_meta_analyses[0].experiment), so we will create one.
            meta_analyzes_experiment_set = meta_analyzes_experiment_sets[0]
            experiment = Experiment(
                experiment_set=meta_analyzes_experiment_set,
                title=item_create.title,
                short_description=item_create.short_description,
                abstract_text=item_create.abstract_text,
                method_text=item_create.method_text,
                extra_metadata={},
                created_by=user_data.user,
                modified_by=user_data.user,
            )
        else:
            experiment = Experiment(
                title=item_create.title,
                short_description=item_create.short_description,
                abstract_text=item_create.abstract_text,
                method_text=item_create.method_text,
                extra_metadata={},
                created_by=user_data.user,
                modified_by=user_data.user,
            )

        save_to_logging_context({"meta_analysis_experiment": experiment.urn})
        logger.debug(msg="Creating experiment within meta analysis experiment.", extra=logging_context())

    contributors: list[Contributor] = []
    try:
        contributors = [
            await find_or_create_contributor(db, contributor.orcid_id) for contributor in item_create.contributors or []
        ]
    except NonexistentOrcidUserError as e:
        logger.error(msg="Could not find ORCID user with the provided user ID.", extra=logging_context())
        raise pydantic.ValidationError(
            [pydantic.error_wrappers.ErrorWrapper(ValidationError(str(e)), loc="contributors")],
            model=score_set.ScoreSetCreate,
        )

    doi_identifiers = [
        await find_or_create_doi_identifier(db, identifier.identifier)
        for identifier in item_create.doi_identifiers or []
    ]
    primary_publication_identifiers = [
        await find_or_create_publication_identifier(db, identifier.identifier, identifier.db_name)
        for identifier in item_create.primary_publication_identifiers or []
    ]
    publication_identifiers = [
        await find_or_create_publication_identifier(db, identifier.identifier, identifier.db_name)
        for identifier in item_create.secondary_publication_identifiers or []
    ] + primary_publication_identifiers

    # create a temporary `primary` attribute on each of our publications that indicates
    # to our association proxy whether it is a primary publication or not
    primary_identifiers = [pub.identifier for pub in primary_publication_identifiers]
    for publication in publication_identifiers:
        setattr(publication, "primary", publication.identifier in primary_identifiers)

    targets: list[TargetGene] = []
    accessions = False
    for gene in item_create.target_genes:
        if gene.target_sequence:
            if accessions and len(targets) > 0:
                logger.info(
                    msg="Failed to create score set; Both a sequence and accession based target were detected.",
                    extra=logging_context(),
                )
                raise MixedTargetError(
                    "MaveDB does not support score-sets with both sequence and accession based targets. Please re-submit this scoreset using only one type of target."
                )
            upload_taxonomy = gene.target_sequence.taxonomy
            save_to_logging_context({"requested_taxonomy": gene.target_sequence.taxonomy.tax_id})
            taxonomy = await find_or_create_taxonomy(db, upload_taxonomy)

            if not taxonomy:
                logger.info(
                    msg="Failed to create score set; The requested taxonomy does not exist.", extra=logging_context()
                )
                raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail="Unknown taxonomy")

            # If the target sequence has a label, use it. Otherwise, use the name from the target gene as the label.
            # View model validation rules enforce that sequences must have a label defined if there are more than one
            # targets defined on a score set.
            seq_label = gene.target_sequence.label if gene.target_sequence.label is not None else gene.name
            target_sequence = TargetSequence(
                **jsonable_encoder(gene.target_sequence, by_alias=False, exclude={"taxonomy", "label"}),
                taxonomy=taxonomy,
                label=seq_label,
            )
            target_gene = TargetGene(
                **jsonable_encoder(
                    gene,
                    by_alias=False,
                    exclude={
                        "external_identifiers",
                        "target_sequence",
                        "target_accession",
                    },
                ),
                target_sequence=target_sequence,
            )

        elif gene.target_accession:
            if not accessions and len(targets) > 0:
                logger.info(
                    msg="Failed to create score set; Both a sequence and accession based target were detected.",
                    extra=logging_context(),
                )

                raise MixedTargetError(
                    "MaveDB does not support score-sets with both sequence and accession based targets. Please re-submit this scoreset using only one type of target."
                )
            accessions = True
            target_accession = TargetAccession(**jsonable_encoder(gene.target_accession, by_alias=False))
            target_gene = TargetGene(
                **jsonable_encoder(
                    gene,
                    by_alias=False,
                    exclude={
                        "external_identifiers",
                        "target_sequence",
                        "target_accession",
                    },
                ),
                target_accession=target_accession,
            )
        else:
            save_to_logging_context({"failing_target": gene})
            logger.info(msg="Failed to create score set; Could not infer target type.", extra=logging_context())
            raise ValueError("One of either `target_accession` or `target_gene` should be present")

        for external_gene_identifier_offset_create in gene.external_identifiers:
            offset = external_gene_identifier_offset_create.offset
            identifier_create = external_gene_identifier_offset_create.identifier
            await create_external_gene_identifier_offset(
                db,
                target_gene,
                identifier_create.db_name,
                identifier_create.identifier,
                offset,
            )

        targets.append(target_gene)

    assert experiment is not None

    item = ScoreSet(
        **jsonable_encoder(
            item_create,
            by_alias=False,
            exclude={
                "contributors",
                "doi_identifiers",
                "experiment_urn",
                "license_id",
                "meta_analyzes_score_set_urns",
                "primary_publication_identifiers",
                "secondary_publication_identifiers",
                "superseded_score_set_urn",
                "target_genes",
                "score_ranges",
            },
        ),
        experiment=experiment,
        license=license_,
        superseded_score_set=superseded_score_set,
        meta_analyzes_score_sets=meta_analyzes_score_sets,
        target_genes=targets,
        contributors=contributors,
        doi_identifiers=doi_identifiers,
        publication_identifiers=publication_identifiers,
        processing_state=ProcessingState.incomplete,
        created_by=user_data.user,
        modified_by=user_data.user,
        score_ranges=item_create.score_ranges.dict() if item_create.score_ranges else null(),
    )  # type: ignore

    db.add(item)
    db.commit()
    db.refresh(item)

    save_to_logging_context({"created_resource": item.urn})
    return item


@router.post(
    "/score-sets/{urn}/variants/data",
    response_model=score_set.ScoreSet,
    responses={422: {}},
    response_model_exclude_none=True,
)
async def upload_score_set_variant_data(
    *,
    urn: str,
    counts_file: Optional[UploadFile] = File(None),
    scores_file: UploadFile = File(...),
    db: Session = Depends(deps.get_db),
    user_data: UserData = Depends(require_current_user_with_email),
    worker: ArqRedis = Depends(deps.get_worker),
) -> Any:
    """
    Upload scores and variant count files for a score set, and initiate processing these files to
    create variants.
    """
    save_to_logging_context({"requested_resource": urn, "resource_property": "variants"})

    # item = db.query(ScoreSet).filter(ScoreSet.urn == urn).filter(ScoreSet.private.is_(False)).one_or_none()
    item = db.query(ScoreSet).filter(ScoreSet.urn == urn).one_or_none()
    if not item or not item.urn:
        logger.info(msg="Failed to create variants; The requested score set does not exist.", extra=logging_context())
        return None

    assert_permission(user_data, item, Action.UPDATE)
    assert_permission(user_data, item, Action.SET_SCORES)

    try:
        scores_df = csv_data_to_df(scores_file.file)
        counts_df = None
        if counts_file and counts_file.filename:
            counts_df = csv_data_to_df(counts_file.file)
    # Handle non-utf8 file problem.
    except UnicodeDecodeError as e:
        raise HTTPException(status_code=400, detail=f"Error decoding file: {e}. Ensure the file has correct values.")

    if scores_file:
        # Although this is also updated within the variant creation job, update it here
        # as well so that we can display the proper UI components (queue invocation delay
        # races the score set GET request).
        item.processing_state = ProcessingState.processing

        # await the insertion of this job into the worker queue, not the job itself.
        job = await worker.enqueue_job(
            "create_variants_for_score_set",
            correlation_id_for_context(),
            item.id,
            user_data.user.id,
            scores_df,
            counts_df,
        )
        if job is not None:
            save_to_logging_context({"worker_job_id": job.job_id})
        logger.info(msg="Enqueud variant creation job.", extra=logging_context())

    db.add(item)
    db.commit()
    db.refresh(item)
    return item


@router.post(
    "/score-sets/{urn}/calibration/data",
    response_model=score_set.ScoreSet,
    responses={422: {}},
    response_model_exclude_none=True,
)
async def update_score_set_calibration_data(
    *,
    urn: str,
    calibration_update: dict[str, calibration.Calibration],
    db: Session = Depends(deps.get_db),
    user_data: UserData = Depends(RoleRequirer([UserRole.admin])),
):
    """
    Update thresholds / score calibrations for a score set.
    """
    save_to_logging_context({"requested_resource": urn, "resource_property": "score_thresholds"})

    try:
        item = db.scalars(select(ScoreSet).where(ScoreSet.urn == urn)).one()
    except NoResultFound:
        logger.info(
            msg="Failed to add score thresholds; The requested score set does not exist.", extra=logging_context()
        )
        raise HTTPException(status_code=404, detail=f"score set with URN '{urn}' not found")

    assert_permission(user_data, item, Action.UPDATE)

    item.score_calibrations = {k: v.dict() for k, v in calibration_update.items()}
    db.add(item)
    db.commit()
    db.refresh(item)

    save_to_logging_context({"updated_resource": item.urn})
    return item


@router.put(
    "/score-sets/{urn}", response_model=score_set.ScoreSet, responses={422: {}}, response_model_exclude_none=True
)
async def update_score_set(
    *,
    urn: str,
    item_update: score_set.ScoreSetUpdate,
    db: Session = Depends(deps.get_db),
    user_data: UserData = Depends(require_current_user_with_email),
    worker: ArqRedis = Depends(deps.get_worker),
) -> Any:
    """
    Update a score set.
    """
    save_to_logging_context({"requested_resource": urn})
    logger.debug(msg="Began score set update.", extra=logging_context())

    item = db.query(ScoreSet).filter(ScoreSet.urn == urn).one_or_none()
    if not item:
        logger.info(msg="Failed to update score set; The requested score set does not exist.", extra=logging_context())
        raise HTTPException(status_code=404, detail=f"score set with URN '{urn}' not found")

    assert_permission(user_data, item, Action.UPDATE)

    for var, value in vars(item_update).items():
        if var not in [
            "contributors",
            "score_ranges",
            "doi_identifiers",
            "experiment_urn",
            "license_id",
            "secondary_publication_identifiers",
            "primary_publication_identifiers",
            "target_genes",
        ]:
            setattr(item, var, value) if value else None

    if item_update.license_id is not None:
        save_to_logging_context({"license": item_update.license_id})
        license_ = db.query(License).filter(License.id == item_update.license_id).one_or_none()

        if not license_:
            logger.info(
                msg="Failed to update score set; The requested license does not exist.", extra=logging_context()
            )
            raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail="Unknown license")

            # Allow in-active licenses to be retained on update if they already exist on the item.
        elif not license_.active and item.licence_id != item_update.license_id:
            logger.info(
                msg="Failed to update score set license; The requested license is no longer active.",
                extra=logging_context(),
            )
            raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail="Invalid license")

        item.license = license_

    item.doi_identifiers = [
        await find_or_create_doi_identifier(db, identifier.identifier)
        for identifier in item_update.doi_identifiers or []
    ]
    primary_publication_identifiers = [
        await find_or_create_publication_identifier(db, identifier.identifier, identifier.db_name)
        for identifier in item_update.primary_publication_identifiers or []
    ]
    publication_identifiers = [
        await find_or_create_publication_identifier(db, identifier.identifier, identifier.db_name)
        for identifier in item_update.secondary_publication_identifiers or []
    ] + primary_publication_identifiers

    # create a temporary `primary` attribute on each of our publications that indicates
    # to our association proxy whether it is a primary publication or not
    primary_identifiers = [p.identifier for p in primary_publication_identifiers]
    for publication in publication_identifiers:
        setattr(publication, "primary", publication.identifier in primary_identifiers)

    item.publication_identifiers = publication_identifiers

    try:
        item.contributors = [
            await find_or_create_contributor(db, contributor.orcid_id) for contributor in item_update.contributors or []
        ]
    except NonexistentOrcidUserError as e:
        logger.error(msg="Could not find ORCID user with the provided user ID.", extra=logging_context())
        raise pydantic.ValidationError(
            [pydantic.error_wrappers.ErrorWrapper(ValidationError(str(e)), loc="contributors")],
            model=score_set.ScoreSetUpdate,
        )

    # Score set has not been published and attributes affecting scores may still be edited.
    if item.private:
        if item_update.score_ranges:
            item.score_ranges = item_update.score_ranges.dict()
        else:
            item.score_ranges = null()

        # Delete the old target gene, WT sequence, and reference map. These will be deleted when we set the score set's
        # target_gene to None, because we have set cascade='all,delete-orphan' on ScoreSet.target_gene. (Since the
        # relationship is defined with the target gene as owner, this is actually set up in the backref attribute of
        # TargetGene.score_set.)
        #
        # We must flush our database queries now so that the old target gene will be deleted before inserting a new one
        # with the same score_set_id.
        item.target_genes = []
        db.flush()

        targets: List[TargetGene] = []
        accessions = False
        for gene in item_update.target_genes:
            if gene.target_sequence:
                if accessions and len(targets) > 0:
                    logger.info(
                        msg="Failed to update score set; Both a sequence and accession based target were detected.",
                        extra=logging_context(),
                    )

                    raise MixedTargetError(
                        "MaveDB does not support score-sets with both sequence and accession based targets. Please re-submit this scoreset using only one type of target."
                    )

                upload_taxonomy = gene.target_sequence.taxonomy
                save_to_logging_context({"requested_taxonomy": gene.target_sequence.taxonomy.tax_id})
                taxonomy = await find_or_create_taxonomy(db, upload_taxonomy)

                if not taxonomy:
                    logger.info(
                        msg="Failed to create score set; The requested taxonomy does not exist.",
                        extra=logging_context(),
                    )
                    raise HTTPException(
                        status_code=status.HTTP_400_BAD_REQUEST,
                        detail=f"Unknown taxonomy {gene.target_sequence.taxonomy.tax_id}",
                    )

                # If the target sequence has a label, use it. Otherwise, use the name from the target gene as the label.
                # View model validation rules enforce that sequences must have a label defined if there are more than one
                # targets defined on a score set.
                seq_label = gene.target_sequence.label if gene.target_sequence.label is not None else gene.name
                target_sequence = TargetSequence(
                    **jsonable_encoder(
                        gene.target_sequence,
                        by_alias=False,
                        exclude={"taxonomy", "label"},
                    ),
                    taxonomy=taxonomy,
                    label=seq_label,
                )
                target_gene = TargetGene(
                    **jsonable_encoder(
                        gene,
                        by_alias=False,
                        exclude={
                            "external_identifiers",
                            "target_sequence",
                            "target_accession",
                        },
                    ),
                    target_sequence=target_sequence,
                )

            elif gene.target_accession:
                if not accessions and len(targets) > 0:
                    logger.info(
                        msg="Failed to create score set; Both a sequence and accession based target were detected.",
                        extra=logging_context(),
                    )
                    raise MixedTargetError(
                        "MaveDB does not support score-sets with both sequence and accession based targets. Please re-submit this scoreset using only one type of target."
                    )
                accessions = True
                target_accession = TargetAccession(**jsonable_encoder(gene.target_accession, by_alias=False))
                target_gene = TargetGene(
                    **jsonable_encoder(
                        gene,
                        by_alias=False,
                        exclude={
                            "external_identifiers",
                            "target_sequence",
                            "target_accession",
                        },
                    ),
                    target_accession=target_accession,
                )
            else:
                save_to_logging_context({"failing_target": gene})
                logger.info(msg="Failed to create score set; Could not infer target type.", extra=logging_context())
                raise ValueError("One of either `target_accession` or `target_gene` should be present")

            for external_gene_identifier_offset_create in gene.external_identifiers:
                offset = external_gene_identifier_offset_create.offset
                identifier_create = external_gene_identifier_offset_create.identifier
                await create_external_gene_identifier_offset(
                    db,
                    target_gene,
                    identifier_create.db_name,
                    identifier_create.identifier,
                    offset,
                )

            targets.append(target_gene)

        item.target_genes = targets

        # re-validate existing variants and clear them if they do not pass validation
        if item.variants:
            assert item.dataset_columns is not None
            score_columns = [
                "hgvs_nt",
                "hgvs_splice",
                "hgvs_pro",
            ] + item.dataset_columns["score_columns"]
            count_columns = [
                "hgvs_nt",
                "hgvs_splice",
                "hgvs_pro",
            ] + item.dataset_columns["count_columns"]

            scores_data = pd.DataFrame(
                variants_to_csv_rows(item.variants, columns=score_columns, dtype="score_data")
            ).replace("NA", pd.NA)

            if item.dataset_columns["count_columns"]:
                count_data = pd.DataFrame(
                    variants_to_csv_rows(item.variants, columns=count_columns, dtype="count_data")
                ).replace("NA", pd.NA)
            else:
                count_data = None

            # Although this is also updated within the variant creation job, update it here
            # as well so that we can display the proper UI components (queue invocation delay
            # races the score set GET request).
            item.processing_state = ProcessingState.processing

            # await the insertion of this job into the worker queue, not the job itself.
            job = await worker.enqueue_job(
                "create_variants_for_score_set",
                correlation_id_for_context(),
                item.id,
                user_data.user.id,
                scores_data,
                count_data,
            )
            if job is not None:
                save_to_logging_context({"worker_job_id": job.job_id})
            logger.info(msg="Enqueud variant creation job.", extra=logging_context())
    else:
        logger.debug(msg="Skipped score range and target gene update. Score set is published.", extra=logging_context())

    db.add(item)
    db.commit()
    db.refresh(item)

    save_to_logging_context({"updated_resource": item.urn})
    return item


@router.delete("/score-sets/{urn}", responses={422: {}})
async def delete_score_set(
    *,
    urn: str,
    db: Session = Depends(deps.get_db),
    user_data: UserData = Depends(require_current_user),
) -> Any:
    """
    Delete a score set.

    Raises

    Returns
    _______
    Does not return anything
    string : HTTP code 200 successful but returning content
    or
    communitcate to client whether the operation succeeded
    204 if successful but not returning content - likely going with this
    """
    save_to_logging_context({"requested_resource": urn})

    item = db.query(ScoreSet).filter(ScoreSet.urn == urn).one_or_none()
    if not item:
        logger.info(msg="Failed to delete score set; The requested score set does not exist.", extra=logging_context())
        raise HTTPException(status_code=404, detail=f"score set with URN '{urn}' not found")

    assert_permission(user_data, item, Action.DELETE)

    db.delete(item)
    db.commit()


@router.post(
    "/score-sets/{urn}/publish",
    status_code=200,
    response_model=score_set.ScoreSet,
    response_model_exclude_none=True,
)
async def publish_score_set(
    *,
    urn: str,
    db: Session = Depends(deps.get_db),
    user_data: UserData = Depends(require_current_user),
    worker: ArqRedis = Depends(deps.get_worker),
) -> Any:
    """
    Publish a score set.
    """
    save_to_logging_context({"requested_resource": urn})

    item: Optional[ScoreSet] = db.query(ScoreSet).filter(ScoreSet.urn == urn).one_or_none()
    if not item:
        logger.info(msg="Failed to publish score set; The requested score set does not exist.", extra=logging_context())
        raise HTTPException(status_code=404, detail=f"score set with URN '{urn}' not found")

    assert_permission(user_data, item, Action.PUBLISH)

    if not item.experiment:
        logger.info(
            msg="Failed to publish score set; The requested score set does not belong to an experiment.",
            extra=logging_context(),
        )
        raise HTTPException(
            status_code=500,
            detail="This score set does not belong to an experiment and cannot be published.",
        )
    if not item.experiment.experiment_set:
        logger.info(
            msg="Failed to publish score set; The requested score set's experiment does not belong to an experiment set.",
            extra=logging_context(),
        )
        raise HTTPException(
            status_code=500,
            detail="This score set's experiment does not belong to an experiment set and cannot be published.",
        )
    # TODO This can probably be done more efficiently; at least, it's worth checking the SQL query that SQLAlchemy
    # generates when all we want is len(score_set.variants).
    if len(item.variants) == 0:
        logger.info(
            msg="Failed to publish score set; The requested score set does not contain any variant scores.",
            extra=logging_context(),
        )
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail="cannot publish score set without variant scores",
        )

    published_date = date.today()

    if item.experiment.experiment_set.private or not item.experiment.experiment_set.published_date:
        item.experiment.experiment_set.urn = generate_experiment_set_urn(db)
        item.experiment.experiment_set.private = False
        item.experiment.experiment_set.published_date = published_date
        db.add(item.experiment.experiment_set)

    save_to_logging_context({"experiment_set": item.experiment.experiment_set.urn})

    if item.experiment.private or not item.experiment.published_date:
        item.experiment.urn = generate_experiment_urn(
            db,
            item.experiment.experiment_set,
            experiment_is_meta_analysis=len(item.meta_analyzes_score_sets) > 0,
        )
        item.experiment.private = False
        item.experiment.published_date = published_date
        db.add(item.experiment)

    save_to_logging_context({"experiment": item.experiment.urn})

    item.urn = generate_score_set_urn(db, item.experiment)
    item.private = False
    item.published_date = published_date
    refresh_variant_urns(db, item)

    save_to_logging_context({"score_set": item.urn})

    db.add(item)
    db.commit()
    db.refresh(item)

    # await the insertion of this job into the worker queue, not the job itself.
    job = await worker.enqueue_job(
        "refresh_published_variants_view",
        correlation_id_for_context(),
        user_data.user.id,
    )
    if job is not None:
        save_to_logging_context({"worker_job_id": job.job_id})
        logger.info(msg="Enqueud published variant materialized view refresh job.", extra=logging_context())
    else:
        logger.warning(
            msg="Failed to enqueue published variant materialized view refresh job.", extra=logging_context()
        )

    return item



import logging
from typing import Any, Optional

from arq import ArqRedis
from arq.jobs import Job
from starlette.requests import Request
from starlette.responses import Response

from mavedb import __version__
from mavedb.lib.logging.context import logging_context, save_to_logging_context
from mavedb.lib.logging.models import LogType, Source

logger = logging.getLogger(__name__)


# NOTE: Starlette context will not be initialized in the worker, so maintain a local dictionary of context. We
#       may eventually want to harden this by using the backing Redis cache or with Python's contextvars.
async def log_job(ctx: dict) -> None:
    redis: ArqRedis = ctx["redis"]
    job_id: str = ctx["job_id"]

    log_context: dict[str, Any] = ctx["state"].pop(job_id) if job_id in ctx["state"] else {}

    completed_job = Job(job_id, redis=redis)
    result = await completed_job.result_info()

    if not result:
        logger.warning(msg=f"Job finished, but could not retrieve a job result for job {job_id}.", extra=log_context)
    else:
        log_context = {
            **log_context,
            **{
                "time_ns": int(result.enqueue_time.timestamp()),
                "queued_ns": (result.start_time - result.enqueue_time).microseconds,
                "duration_ns": (result.finish_time - result.start_time).microseconds,
                "job_name": result.function,
                "job_attempt": result.job_try,
                "arq_success": result.success,
                "job_result": result.result,
            },
        }

    log_context = {
        **log_context,
        **{
            "job_id": completed_job.job_id,
            "version": __version__,
            "log_type": LogType.worker_job,
            "source": Source.worker,
            "canonical": True,
        },
    }

    if result is None:
        logger.error(msg="Job result could not be found.", extra=log_context)
    elif result.result is not None:
        logger.info(msg="Job completed with result.", extra=log_context)
    else:
        logger.error(msg="Job completed with unhandled exception.", extra=log_context)


def log_request(request: Request, response: Response, end: int) -> None:
    save_to_logging_context({"log_type": LogType.api_request, "response_code": response.status_code})

    start: Optional[int] = logging_context().get("time_ns")
    if start:
        save_to_logging_context({"duration_ns": end - start})

    save_to_logging_context({"canonical": True})
    if response.status_code < 400:
        logger.info(msg="Request completed.", extra=logging_context())
    elif response.status_code < 500:
        logger.warning(msg="Request completed.", extra=logging_context())
    else:
        logger.error(msg="Request completed.", extra=logging_context())



from datetime import date

from sqlalchemy import Boolean, Column, Date, Integer, String, UniqueConstraint

from mavedb.db.base import Base


class ControlledKeyword(Base):
    __tablename__ = "controlled_keywords"

    id = Column(Integer, primary_key=True, index=True)
    key = Column(String, nullable=False)
    value = Column(String, nullable=False)
    vocabulary = Column(String, nullable=True)
    special = Column(Boolean, nullable=True)
    description = Column(String, nullable=True)
    creation_date = Column(Date, nullable=False, default=date.today)
    modification_date = Column(Date, nullable=False, default=date.today, onupdate=date.today)
    __table_args__ = (UniqueConstraint("key", "value", name="ix_controlled_keywords_key_value"),)



import logging
from datetime import date
from typing import Any, Dict, Optional, Sequence

from fastapi import APIRouter, Depends, HTTPException
from fastapi.encoders import jsonable_encoder
from sqlalchemy import and_, select
from sqlalchemy.orm import Session
from sqlalchemy.orm.exc import MultipleResultsFound, NoResultFound

from mavedb import deps
from mavedb.lib.authentication import UserData, get_current_user
from mavedb.lib.authorization import require_current_user, require_current_user_with_email
from mavedb.lib.logging import LoggedRoute
from mavedb.lib.logging.context import (
    format_raised_exception_info_as_dict,
    logging_context,
    save_to_logging_context,
)
from mavedb.lib.permissions import Action, assert_permission, has_permission
from mavedb.models.collection import Collection
from mavedb.models.collection_user_association import CollectionUserAssociation
from mavedb.models.enums.contribution_role import ContributionRole
from mavedb.models.experiment import Experiment
from mavedb.models.score_set import ScoreSet
from mavedb.models.user import User
from mavedb.view_models import collection
from mavedb.view_models import collection_bundle

logger = logging.getLogger(__name__)

router = APIRouter(
    prefix="/api/v1",
    tags=["collections"],
    responses={404: {"description": "Not found"}},
    route_class=LoggedRoute,
)


@router.get(
    "/users/me/collections",
    status_code=200,
    response_model=collection_bundle.CollectionBundle,
    response_model_exclude_none=True,
)
def list_my_collections(
    *,
    db: Session = Depends(deps.get_db),
    user_data: UserData = Depends(require_current_user),
) -> Dict[str, Sequence[Collection]]:
    """
    List my collections.
    """
    collection_bundle: Dict[str, Sequence[Collection]] = {}
    for role in ContributionRole:
        collection_bundle[role.value] = (
            db.execute(
                select(Collection)
                .join(CollectionUserAssociation)
                .where(CollectionUserAssociation.user_id == user_data.user.id)
                .where(CollectionUserAssociation.contribution_role == role.value)
            )
            .scalars()
            .all()
        )

        for item in collection_bundle[role.value]:
            # filter score sets and experiments based on user permissions
            item.score_sets = [
                score_set for score_set in item.score_sets if has_permission(user_data, score_set, Action.READ)
            ]
            item.experiments = [
                experiment for experiment in item.experiments if has_permission(user_data, experiment, Action.READ)
            ]
            # unless user is admin of this collection, filter users to only admins
            # the rationale is that all collection contributors should be able to see admins
            # to know who to contact, but only collection admins should be able to see viewers and editors
            if role in (ContributionRole.viewer, ContributionRole.editor):
                admins = []
                for user_assoc in item.user_associations:
                    if user_assoc.contribution_role == ContributionRole.admin:
                        admin = user_assoc.user
                        # role must be set in order to assign users to collection
                        setattr(admin, "role", ContributionRole.admin)
                        admins.append(admin)
                item.users = admins

    return collection_bundle


@router.get(
    "/collections/{urn}",
    status_code=200,
    response_model=collection.Collection,
    responses={404: {}},
    response_model_exclude_none=True,
)
def fetch_collection(
    *,
    urn: str,
    db: Session = Depends(deps.get_db),
    user_data: Optional[UserData] = Depends(get_current_user),
) -> Collection:
    """
    Fetch a single collection by URN.
    """
    save_to_logging_context({"requested_resource": urn})

    item = db.execute(select(Collection).where(Collection.urn == urn)).scalars().one_or_none()
    if not item:
        logger.debug(msg="The requested collection does not exist.", extra=logging_context())
        raise HTTPException(status_code=404, detail=f"Collection with URN {urn} not found")

    assert_permission(user_data, item, Action.READ)
    # filter score sets and experiments based on user permissions
    item.score_sets = [score_set for score_set in item.score_sets if has_permission(user_data, score_set, Action.READ)]
    item.experiments = [
        experiment for experiment in item.experiments if has_permission(user_data, experiment, Action.READ)
    ]

    # Only collection admins can see all user roles for the collection. Other users can only see the list of admins.
    # We could create a new permission action for this. But for now, assume that any user who has the ADD_ROLE
    # permission is a collection admin and should be able to see all user roles for the collection.
    if not has_permission(user_data, item, Action.ADD_ROLE):
        admins = []
        for user_assoc in item.user_associations:
            if user_assoc.contribution_role == ContributionRole.admin:
                admin = user_assoc.user
                # role must be set in order to assign users to collection
                setattr(admin, "role", ContributionRole.admin)
                admins.append(admin)
        item.users = admins

    return item


@router.post(
    "/collections/",
    response_model=collection.Collection,
    responses={422: {}},
    response_model_exclude_none=True,
)
async def create_collection(
    *,
    item_create: collection.CollectionCreate,
    db: Session = Depends(deps.get_db),
    user_data: UserData = Depends(require_current_user_with_email),
) -> Any:
    """
    Create a collection.
    """
    logger.debug(msg="Began creation of new collection.", extra=logging_context())

    users = []
    user_orcid_ids = set()

    try:
        # always assign creator as admin, as collections permissions do not distinguish between owner/creator and admin
        creator_user = user_data.user
        setattr(creator_user, "role", ContributionRole.admin)
        users.append(creator_user)
        user_orcid_ids.add(creator_user.username)

        for admin in item_create.admins or []:
            admin_orcid = admin.orcid_id
            if admin_orcid not in user_orcid_ids:
                user = db.scalars(select(User).where(User.username == admin_orcid)).one()
                setattr(user, "role", ContributionRole.admin)
                users.append(user)
                user_orcid_ids.add(admin_orcid)

        for editor in item_create.editors or []:
            editor_orcid = editor.orcid_id
            if editor_orcid not in user_orcid_ids:
                user = db.scalars(select(User).where(User.username == editor_orcid)).one()
                setattr(user, "role", ContributionRole.editor)
                users.append(user)
                user_orcid_ids.add(editor_orcid)

        for viewer in item_create.viewers or []:
            viewer_orcid = viewer.orcid_id
            if viewer_orcid not in user_orcid_ids:
                user = db.scalars(select(User).where(User.username == viewer_orcid)).one()
                setattr(user, "role", ContributionRole.viewer)
                users.append(user)
                user_orcid_ids.add(viewer_orcid)

    except NoResultFound as e:
        save_to_logging_context(format_raised_exception_info_as_dict(e))
        logger.error(
            msg="No existing user found with the given ORCID iD",
            extra=logging_context(),
        )
        raise HTTPException(status_code=404, detail="No MaveDB user found with the given ORCID iD")

    except MultipleResultsFound as e:
        save_to_logging_context(format_raised_exception_info_as_dict(e))
        logger.error(msg="Multiple users found with the given ORCID iD", extra=logging_context())
        raise HTTPException(
            status_code=400,
            detail="Multiple MaveDB users found with the given ORCID iD",
        )

    try:
        score_sets = [
            db.scalars(select(ScoreSet).where(ScoreSet.urn == score_set_urn)).one()
            for score_set_urn in item_create.score_set_urns or []
        ]

        experiments = [
            db.scalars(select(Experiment).where(Experiment.urn == experiment_urn)).one()
            for experiment_urn in item_create.experiment_urns or []
        ]

    except NoResultFound as e:
        save_to_logging_context(format_raised_exception_info_as_dict(e))
        logger.error(msg="No resource found with the given URN", extra=logging_context())
        raise HTTPException(status_code=404, detail="No resource found with the given URN")

    except MultipleResultsFound as e:
        save_to_logging_context(format_raised_exception_info_as_dict(e))
        logger.error(msg="Multiple resources found with the given URN", extra=logging_context())
        raise HTTPException(status_code=400, detail="Multiple resources found with the given URN")

    item = Collection(
        **jsonable_encoder(
            item_create,
            by_alias=False,
            exclude={
                "viewers",
                "editors",
                "admins",
                "score_set_urns",
                "experiment_urns",
                "badge_name",
            },
        ),
        users=users,
        score_sets=score_sets,
        experiments=experiments,
        created_by=user_data.user,
        modified_by=user_data.user,
    )  # type: ignore

    db.add(item)
    db.commit()
    db.refresh(item)

    save_to_logging_context({"created_resource": item.urn})
    return item


@router.patch(
    "/collections/{urn}",
    response_model=collection.Collection,
    responses={422: {}},
    response_model_exclude_none=True,
)
async def update_collection(
    *,
    item_update: collection.CollectionModify,
    urn: str,
    db: Session = Depends(deps.get_db),
    user_data: UserData = Depends(require_current_user_with_email),
) -> Any:
    """
    Modify a collection's metadata.
    """
    save_to_logging_context({"requested_resource": urn})
    logger.debug(msg="Began collection metadata update.", extra=logging_context())

    item = db.execute(select(Collection).where(Collection.urn == urn)).scalars().one_or_none()
    if item is None:
        logger.info(
            msg="Failed to update collection; The requested collection does not exist.",
            extra=logging_context(),
        )
        raise HTTPException(status_code=404, detail=f"collection with URN {urn} not found")

    assert_permission(user_data, item, Action.UPDATE)

    # Editors may update metadata, but not all editors can publish (which is just setting private to public).
    if item.private and not item_update.private:
        assert_permission(user_data, item, Action.PUBLISH)

    # Unpublishing requires the same permissions as publishing.
    if not item.private and item_update.private:
        assert_permission(user_data, item, Action.PUBLISH)

    if item_update.badge_name:
        assert_permission(user_data, item, Action.ADD_BADGE)

    # Only access fields set by the user. Note the value of set fields will be updated even if the value is None
    pairs = {k: v for k, v in vars(item_update).items() if k in item_update.__fields_set__}
    for var, value in pairs.items():  # vars(item_update).items():
        setattr(item, var, value)

    item.modified_by = user_data.user

    db.add(item)
    db.commit()
    db.refresh(item)

    save_to_logging_context({"updated_resource": item.urn})
    # filter score sets and experiments based on user permissions
    # note that this filtering occurs after saving changes to db; the filtering is only for the returned view model
    item.score_sets = [score_set for score_set in item.score_sets if has_permission(user_data, score_set, Action.READ)]
    item.experiments = [
        experiment for experiment in item.experiments if has_permission(user_data, experiment, Action.READ)
    ]

    # Only collection admins can see all user roles for the collection. Other users can only see the list of admins.
    # We could create a new permission action for this. But for now, assume that any user who has the ADD_ROLE
    # permission is a collection admin and should be able to see all user roles for the collection.
    if not has_permission(user_data, item, Action.ADD_ROLE):
        admins = []
        for user_assoc in item.user_associations:
            if user_assoc.contribution_role == ContributionRole.admin:
                admin = user_assoc.user
                # role must be set in order to assign users to collection
                setattr(admin, "role", ContributionRole.admin)
                admins.append(admin)
        item.users = admins

    return item


@router.post(
    "/collections/{collection_urn}/score-sets",
    response_model=collection.Collection,
    responses={422: {}},
)
async def add_score_set_to_collection(
    *,
    body: collection.AddScoreSetToCollectionRequest,
    collection_urn: str,
    db: Session = Depends(deps.get_db),
    user_data: UserData = Depends(require_current_user_with_email),
) -> Any:
    """
    Add an existing score set to an existing collection.
    """
    save_to_logging_context({"requested_resource": collection_urn})

    item = db.execute(select(Collection).where(Collection.urn == collection_urn)).scalars().one_or_none()
    if not item:
        logger.info(
            msg="Failed to add score set to collection; The requested collection does not exist.",
            extra=logging_context(),
        )
        raise HTTPException(status_code=404, detail=f"collection with URN '{collection_urn}' not found")

    score_set = db.execute(select(ScoreSet).where(ScoreSet.urn == body.score_set_urn)).scalars().one_or_none()
    if not score_set:
        logger.info(
            msg="Failed to add score set to collection; The requested score set does not exist.",
            extra=logging_context(),
        )
        raise HTTPException(
            status_code=404,
            detail=f"score set with URN '{body.score_set_urn}' not found",
        )

    assert_permission(user_data, item, Action.ADD_SCORE_SET)

    item.score_sets.append(score_set)
    item.modification_date = date.today()
    item.modified_by = user_data.user

    db.add(item)
    db.commit()
    db.refresh(item)

    save_to_logging_context({"updated_resource": item.urn})

    # filter score sets and experiments based on user permissions
    # note that this filtering occurs after saving changes to db; the filtering is only for the returned view model
    item.score_sets = [score_set for score_set in item.score_sets if has_permission(user_data, score_set, Action.READ)]
    item.experiments = [
        experiment for experiment in item.experiments if has_permission(user_data, experiment, Action.READ)
    ]

    # Only collection admins can see all user roles for the collection. Other users can only see the list of admins.
    # We could create a new permission action for this. But for now, assume that any user who has the ADD_ROLE
    # permission is a collection admin and should be able to see all user roles for the collection.
    if not has_permission(user_data, item, Action.ADD_ROLE):
        admins = []
        for user_assoc in item.user_associations:
            if user_assoc.contribution_role == ContributionRole.admin:
                admin = user_assoc.user
                # role must be set in order to assign users to collection
                setattr(admin, "role", ContributionRole.admin)
                admins.append(admin)
        item.users = admins

    return item


@router.delete(
    "/collections/{collection_urn}/score-sets/{score_set_urn}",
    response_model=collection.Collection,
    responses={422: {}},
)
async def delete_score_set_from_collection(
    *,
    collection_urn: str,
    score_set_urn: str,
    db: Session = Depends(deps.get_db),
    user_data: UserData = Depends(require_current_user_with_email),
) -> Any:
    """
    Remove a score set from an existing collection. Preserves the score set in the database, only removes the association between the score set and the collection.
    """
    save_to_logging_context({"requested_resource": collection_urn})

    item = db.execute(select(Collection).where(Collection.urn == collection_urn)).scalars().one_or_none()
    if not item:
        logger.info(
            msg="Failed to remove score set from collection; The requested collection does not exist.",
            extra=logging_context(),
        )
        raise HTTPException(status_code=404, detail=f"collection with URN '{collection_urn}' not found")

    score_set = db.execute(select(ScoreSet).where(ScoreSet.urn == score_set_urn)).scalars().one_or_none()
    if not score_set:
        logger.info(
            msg="Failed to remove score set from collection; The requested score set does not exist.",
            extra=logging_context(),
        )
        raise HTTPException(status_code=404, detail=f"score set with URN '{score_set_urn}' not found")

    if score_set not in item.score_sets:
        logger.info(
            msg="Failed to remove score set from collection; The requested score set is not associated with the requested collection.",
            extra=logging_context(),
        )
        raise HTTPException(
            status_code=404,
            detail=f"association between score set '{score_set_urn}' and collection '{collection_urn}' not found",
        )

    # add and remove permissions are the same
    assert_permission(user_data, item, Action.ADD_SCORE_SET)

    item.score_sets.remove(score_set)
    item.modification_date = date.today()
    item.modified_by = user_data.user

    db.add(item)
    db.commit()
    db.refresh(item)

    save_to_logging_context({"updated_resource": item.urn})

    # filter score sets and experiments based on user permissions
    # note that this filtering occurs after saving changes to db; the filtering is only for the returned view model
    item.score_sets = [score_set for score_set in item.score_sets if has_permission(user_data, score_set, Action.READ)]
    item.experiments = [
        experiment for experiment in item.experiments if has_permission(user_data, experiment, Action.READ)
    ]

    # Only collection admins can see all user roles for the collection. Other users can only see the list of admins.
    # We could create a new permission action for this. But for now, assume that any user who has the ADD_ROLE
    # permission is a collection admin and should be able to see all user roles for the collection.
    if not has_permission(user_data, item, Action.ADD_ROLE):
        admins = []
        for user_assoc in item.user_associations:
            if user_assoc.contribution_role == ContributionRole.admin:
                admin = user_assoc.user
                # role must be set in order to assign users to collection
                setattr(admin, "role", ContributionRole.admin)
                admins.append(admin)
        item.users = admins

    return item


@router.post(
    "/collections/{collection_urn}/experiments",
    response_model=collection.Collection,
    responses={422: {}},
)
async def add_experiment_to_collection(
    *,
    body: collection.AddExperimentToCollectionRequest,
    collection_urn: str,
    db: Session = Depends(deps.get_db),
    user_data: UserData = Depends(require_current_user_with_email),
) -> Any:
    """
    Add an existing experiment to an existing collection.
    """
    save_to_logging_context({"requested_resource": collection_urn})

    item = db.execute(select(Collection).where(Collection.urn == collection_urn)).scalars().one_or_none()
    if not item:
        logger.info(
            msg="Failed to add experiment to collection; The requested collection does not exist.",
            extra=logging_context(),
        )
        raise HTTPException(status_code=404, detail=f"collection with URN '{collection_urn}' not found")

    experiment = db.execute(select(Experiment).where(Experiment.urn == body.experiment_urn)).scalars().one_or_none()
    if not experiment:
        logger.info(
            msg="Failed to add experiment to collection; The requested experiment does not exist.",
            extra=logging_context(),
        )
        raise HTTPException(
            status_code=404,
            detail=f"experiment with URN '{body.experiment_urn}' not found",
        )

    assert_permission(user_data, item, Action.ADD_EXPERIMENT)

    item.experiments.append(experiment)
    item.modification_date = date.today()
    item.modified_by = user_data.user

    db.add(item)
    db.commit()
    db.refresh(item)

    save_to_logging_context({"updated_resource": item.urn})

    # filter score sets and experiments based on user permissions
    # note that this filtering occurs after saving changes to db; the filtering is only for the returned view model
    item.score_sets = [score_set for score_set in item.score_sets if has_permission(user_data, score_set, Action.READ)]
    item.experiments = [
        experiment for experiment in item.experiments if has_permission(user_data, experiment, Action.READ)
    ]

    # Only collection admins can see all user roles for the collection. Other users can only see the list of admins.
    # We could create a new permission action for this. But for now, assume that any user who has the ADD_ROLE
    # permission is a collection admin and should be able to see all user roles for the collection.
    if not has_permission(user_data, item, Action.ADD_ROLE):
        admins = []
        for user_assoc in item.user_associations:
            if user_assoc.contribution_role == ContributionRole.admin:
                admin = user_assoc.user
                # role must be set in order to assign users to collection
                setattr(admin, "role", ContributionRole.admin)
                admins.append(admin)
        item.users = admins

    return item


@router.delete(
    "/collections/{collection_urn}/experiments/{experiment_urn}",
    response_model=collection.Collection,
    responses={422: {}},
)
async def delete_experiment_from_collection(
    *,
    collection_urn: str,
    experiment_urn: str,
    db: Session = Depends(deps.get_db),
    user_data: UserData = Depends(require_current_user_with_email),
) -> Any:
    """
    Remove an experiment from an existing collection. Preserves the experiment in the database, only removes the association between the experiment and the collection.
    """
    save_to_logging_context({"requested_resource": collection_urn})

    item = db.execute(select(Collection).where(Collection.urn == collection_urn)).scalars().one_or_none()
    if not item:
        logger.info(
            msg="Failed to remove experiment from collection; The requested collection does not exist.",
            extra=logging_context(),
        )
        raise HTTPException(status_code=404, detail=f"collection with URN '{collection_urn}' not found")

    experiment = db.execute(select(Experiment).where(Experiment.urn == experiment_urn)).scalars().one_or_none()
    if not experiment:
        logger.info(
            msg="Failed to remove experiment from collection; The requested experiment does not exist.",
            extra=logging_context(),
        )
        raise HTTPException(status_code=404, detail=f"experiment with URN '{experiment_urn}' not found")

    if experiment not in item.experiments:
        logger.info(
            msg="Failed to remove experiment from collection; The requested experiment is not associated with the requested collection.",
            extra=logging_context(),
        )
        raise HTTPException(
            status_code=404,
            detail=f"association between experiment '{experiment_urn}' and collection '{collection_urn}' not found",
        )

    # add and remove permissions are the same
    assert_permission(user_data, item, Action.ADD_EXPERIMENT)

    item.experiments.remove(experiment)
    item.modification_date = date.today()
    item.modified_by = user_data.user

    db.add(item)
    db.commit()
    db.refresh(item)

    save_to_logging_context({"updated_resource": item.urn})

    # filter score sets and experiments based on user permissions
    # note that this filtering occurs after saving changes to db; the filtering is only for the returned view model
    item.score_sets = [score_set for score_set in item.score_sets if has_permission(user_data, score_set, Action.READ)]
    item.experiments = [
        experiment for experiment in item.experiments if has_permission(user_data, experiment, Action.READ)
    ]

    # Only collection admins can see all user roles for the collection. Other users can only see the list of admins.
    # We could create a new permission action for this. But for now, assume that any user who has the ADD_ROLE
    # permission is a collection admin and should be able to see all user roles for the collection.
    if not has_permission(user_data, item, Action.ADD_ROLE):
        admins = []
        for user_assoc in item.user_associations:
            if user_assoc.contribution_role == ContributionRole.admin:
                admin = user_assoc.user
                # role must be set in order to assign users to collection
                setattr(admin, "role", ContributionRole.admin)
                admins.append(admin)
        item.users = admins

    return item


@router.post(
    "/collections/{urn}/{role}s",
    response_model=collection.Collection,
    responses={422: {}},
)
async def add_user_to_collection_role(
    *,
    body: collection.AddUserToCollectionRoleRequest,
    urn: str,
    role: ContributionRole,
    db: Session = Depends(deps.get_db),
    user_data: UserData = Depends(require_current_user_with_email),
) -> Any:
    """
    Add an existing user to a collection under the specified role.
    Removes the user from any other roles in this collection.
    """
    save_to_logging_context({"requested_resource": urn})

    item = db.execute(select(Collection).where(Collection.urn == urn)).scalars().one_or_none()
    if not item:
        logger.info(
            msg="Failed to add user to collection role; The requested collection does not exist.",
            extra=logging_context(),
        )
        raise HTTPException(status_code=404, detail=f"collection with URN '{urn}' not found")

    user = db.execute(select(User).where(User.username == body.orcid_id)).scalars().one_or_none()
    if not user:
        logger.info(
            msg="Failed to add user to collection role; The requested user does not exist.",
            extra=logging_context(),
        )
        raise HTTPException(status_code=404, detail=f"user with ORCID iD '{body.orcid_id}' not found")

    # get current user role
    collection_user_association = (
        db.execute(
            select(CollectionUserAssociation)
            .where(CollectionUserAssociation.collection_id == item.id)
            .where(CollectionUserAssociation.user_id == user.id)
        )
        .scalars()
        .one_or_none()
    )

    assert_permission(user_data, item, Action.ADD_ROLE)

    # Since this is a post request, user should not already be in this role
    if collection_user_association and collection_user_association.contribution_role == role:
        logger.info(
            msg="Failed to add user to collection role; the requested user already has the requested role for this collection.",
            extra=logging_context(),
        )
        raise HTTPException(
            status_code=400,
            detail=f"user with ORCID iD '{body.orcid_id}' is already a {role} for collection '{urn}'",
        )
    # A user can only be in one role per collection, so remove from any other roles
    elif collection_user_association:
        item.users.remove(user)

    setattr(user, "role", role)
    item.users.append(user)

    item.modified_by = user_data.user

    db.add(item)
    db.commit()
    db.refresh(item)

    save_to_logging_context({"updated_resource": item.urn})

    # filter score sets and experiments based on user permissions
    # note that this filtering occurs after saving changes to db; the filtering is only for the returned view model
    item.score_sets = [score_set for score_set in item.score_sets if has_permission(user_data, score_set, Action.READ)]
    item.experiments = [
        experiment for experiment in item.experiments if has_permission(user_data, experiment, Action.READ)
    ]

    # Only collection admins can get to this point in the function, so here we don't need to filter the list of user
    # roles to show only admins.

    return item


@router.delete(
    "/collections/{urn}/{role}s/{orcid_id}",
    response_model=collection.Collection,
    responses={422: {}},
)
async def remove_user_from_collection_role(
    *,
    urn: str,
    role: ContributionRole,
    orcid_id: str,
    db: Session = Depends(deps.get_db),
    user_data: UserData = Depends(require_current_user_with_email),
) -> Any:
    """
    Remove a user from a collection role.
    """
    save_to_logging_context({"requested_resource": urn})

    item = db.execute(select(Collection).where(Collection.urn == urn)).scalars().one_or_none()
    if not item:
        logger.info(
            msg="Failed to add user to collection role; The requested collection does not exist.",
            extra=logging_context(),
        )
        raise HTTPException(status_code=404, detail=f"collection with URN '{urn}' not found")

    user = db.execute(select(User).where(User.username == orcid_id)).scalars().one_or_none()
    if not user:
        logger.info(
            msg="Failed to add user to collection role; The requested user does not exist.",
            extra=logging_context(),
        )
        raise HTTPException(status_code=404, detail=f"user with ORCID iD '{orcid_id}' not found")

    # get current user role
    collection_user_association = (
        db.execute(
            select(CollectionUserAssociation).where(
                and_(
                    CollectionUserAssociation.collection_id == item.id,
                    CollectionUserAssociation.user_id == user.id,
                )
            )
        )
        .scalars()
        .one_or_none()
    )

    assert_permission(user_data, item, Action.ADD_ROLE)

    # Since this is a post request, user should not already be in this role
    if collection_user_association is not None and collection_user_association.contribution_role != role:
        logger.info(
            msg="Failed to remove user from collection role; the requested user does not currently hold the requested role for this collection.",
            extra=logging_context(),
        )
        raise HTTPException(
            status_code=404,
            detail=f"user with ORCID iD '{orcid_id}' does not currently hold the role {role} for collection '{urn}'",
        )

    item.users.remove(user)
    item.modified_by = user_data.user

    db.add(item)
    db.commit()
    db.refresh(item)

    save_to_logging_context({"updated_resource": item.urn})

    # filter score sets and experiments based on user permissions
    # note that this filtering occurs after saving changes to db; the filtering is only for the returned view model
    item.score_sets = [score_set for score_set in item.score_sets if has_permission(user_data, score_set, Action.READ)]
    item.experiments = [
        experiment for experiment in item.experiments if has_permission(user_data, experiment, Action.READ)
    ]

    # Only collection admins can get to this point in the function, so here we don't need to filter the list of user
    # roles to show only admins.

    return item


@router.delete("/collections/{urn}", responses={422: {}})
async def delete_collection(
    *,
    urn: str,
    db: Session = Depends(deps.get_db),
    user_data: UserData = Depends(require_current_user_with_email),
) -> Any:
    """
    Delete a collection.
    """
    save_to_logging_context({"requested_resource": urn})

    item = db.execute(select(Collection).where(Collection.urn == urn)).scalars().one_or_none()
    if not item:
        logger.info(
            msg="Failed to delete collection; The requested collection does not exist.",
            extra=logging_context(),
        )
        raise HTTPException(status_code=404, detail=f"collection with URN '{urn}' not found")

    assert_permission(user_data, item, Action.DELETE)

    db.delete(item)
    db.commit()



from typing import Optional

from sqlalchemy import func
from sqlalchemy.orm import Session

from mavedb.models.controlled_keyword import ControlledKeyword


def search_keyword(db: Session, key: str, value: Optional[str]):
    lower_key = key.lower().strip()
    lower_value = value.lower().strip() if value is not None else None
    query = db.query(ControlledKeyword)
    if lower_key:
        query = query.filter(func.lower(ControlledKeyword.key) == lower_key)
    if lower_value:
        query = query.filter(func.lower(ControlledKeyword.value) == lower_value)

    controlled_keyword = query.one_or_none()
    if controlled_keyword is None:
        raise ValueError(f"Invalid keyword {key} or {value}")
    return controlled_keyword



from typing import Any

from fastapi import APIRouter

from mavedb.lib.logging import LoggedRoute
from mavedb.lib.logging.context import logging_context, save_to_logging_context

router = APIRouter(
    prefix="/api/v1/log",
    tags=["log"],
    responses={404: {"description": "Not found"}},
    route_class=LoggedRoute,
)


# NOTE: Despite not containing any calls to a logger, this route will log posted context
#       by nature of its inheritance from LoggedRoute.
@router.post("/", status_code=200, response_model=str, responses={404: {}})
def log_it(log_context: dict) -> Any:
    """
    Log an interaction.
    """
    # Overwrites middleware generated context with context from POST request. It
    # may be that the posted log context contains some colliding state (which we
    # should treat preferentially, so as to retain any previously generated state)
    # to that generated by the middleware.
    ctx = {**logging_context(), **log_context}
    ctx = save_to_logging_context(ctx)

    return ctx.get("X-Correlation-ID")



from datetime import date
from typing import Any, Sequence

from pydantic import Field
from pydantic.types import Optional

from mavedb.view_models import UserContributionRoleGetter, record_type_validator, set_record_type
from mavedb.view_models.base.base import BaseModel
from mavedb.view_models.contributor import ContributorCreate
from mavedb.view_models.user import SavedUser, User


class CollectionGetter(UserContributionRoleGetter):
    def get(self, key: Any, default: Any = ...) -> Any:
        if key == "score_set_urns":
            score_sets = getattr(self._obj, "score_sets") or []
            return sorted([score_set.urn for score_set in score_sets if score_set.superseding_score_set is None])
        elif key == "experiment_urns":
            experiments = getattr(self._obj, "experiments") or []
            return sorted([experiment.urn for experiment in experiments])
        else:
            return super().get(key, default)


class CollectionBase(BaseModel):
    private: bool = Field(
        description="Whether the collection is visible to all MaveDB users. If set during collection update, input ignored unless requesting user is collection admin."
    )
    name: str
    description: Optional[str]
    badge_name: Optional[str] = Field(
        description="Badge name. Input ignored unless requesting user has MaveDB admin privileges."
    )


class CollectionModify(BaseModel):
    # all fields should be optional, because the client should specify only the fields they want to update
    private: Optional[bool] = Field(
        description="Whether the collection is visible to all MaveDB users. If set during collection update, input ignored unless requesting user is collection admin."
    )
    name: Optional[str]
    description: Optional[str]
    badge_name: Optional[str] = Field(
        description="Badge name. Input ignored unless requesting user has MaveDB admin privileges."
    )


class CollectionCreate(CollectionBase):
    experiment_urns: Optional[list[str]]
    score_set_urns: Optional[list[str]]

    viewers: Optional[list[ContributorCreate]]
    editors: Optional[list[ContributorCreate]]
    admins: Optional[list[ContributorCreate]]


class AddScoreSetToCollectionRequest(BaseModel):
    score_set_urn: str


class AddExperimentToCollectionRequest(BaseModel):
    experiment_urn: str


class AddUserToCollectionRoleRequest(BaseModel):
    orcid_id: str


# Properties shared by models stored in DB
class SavedCollection(CollectionBase):
    record_type: str = None  # type: ignore
    urn: str

    created_by: Optional[SavedUser]
    modified_by: Optional[SavedUser]

    experiment_urns: list[str]
    score_set_urns: list[str]

    admins: Sequence[SavedUser]
    viewers: Sequence[SavedUser]
    editors: Sequence[SavedUser]

    creation_date: date
    modification_date: date

    _record_type_factory = record_type_validator()(set_record_type)

    class Config:
        orm_mode = True
        getter_dict = CollectionGetter


# Properties to return to non-admin clients
# NOTE: Coupled to ContributionRole enum
class Collection(SavedCollection):
    created_by: Optional[User]
    modified_by: Optional[User]

    admins: Sequence[User]
    viewers: Sequence[User]
    editors: Sequence[User]


# Properties to return to admin clients or non-admin clients who are admins of the returned collection
# NOTE: Coupled to ContributionRole enum
class AdminCollection(Collection):
    pass



from sqlalchemy import select, join

from mavedb.db.view import MaterializedView, view

from mavedb.models.score_set import ScoreSet
from mavedb.models.variant import Variant
from mavedb.models.mapped_variant import MappedVariant


signature = "published_variants_materialized_view"
definition = (
    select(
        Variant.id.label("variant_id"),
        Variant.urn.label("variant_urn"),
        MappedVariant.id.label("mapped_variant_id"),
        ScoreSet.id.label("score_set_id"),
        ScoreSet.urn.label("score_set_urn"),
        ScoreSet.published_date.label("published_date"),
        MappedVariant.current.label("current_mapped_variant"),
    )
    .select_from(
        join(Variant, MappedVariant, Variant.id == MappedVariant.variant_id, isouter=True).join(
            ScoreSet, ScoreSet.id == Variant.score_set_id
        )
    )
    .where(
        ScoreSet.published_date.is_not(None),
    )
)


class PublishedVariantsMV(MaterializedView):
    __table__ = view(
        signature,
        definition,
        materialized=True,
    )

    variant_id = __table__.c.variant_id
    variant_urn = __table__.c.variant_urn
    mapped_variant_id = __table__.c.mapped_variant_id
    score_set_id = __table__.c.score_set_id
    score_set_urn = __table__.c.score_set_urn
    published_date = __table__.c.published_date
    current_mapped_variant = __table__.c.current_mapped_variant



import json
import logging
import requests
from typing import Sequence, Optional

import click
from sqlalchemy import select
from sqlalchemy.orm import Session

from mavedb.models.score_set import ScoreSet
from mavedb.models.mapped_variant import MappedVariant
from mavedb.models.target_gene import TargetGene
from mavedb.models.variant import Variant

from mavedb.scripts.environment import script_environment, with_database_session

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

CLINGEN_API_URL = "https://reg.test.genome.network/allele"


def get_gene_symbol_from_clingen(hgvs_string: str) -> Optional[str]:
    response = requests.get(f"{CLINGEN_API_URL}?hgvs={hgvs_string}")
    if response.status_code != 200:
        logger.error(f"Failed to query ClinGen API for {hgvs_string}: {response.status_code}")
        return None

    data = response.json()
    if "aminoAcidAlleles" in data:
        return data["aminoAcidAlleles"][0].get("geneSymbol")
    elif "transcriptAlleles" in data:
        return data["transcriptAlleles"][0].get("geneSymbol")

    return None


@script_environment.command()
@with_database_session
@click.argument("urns", nargs=-1)
@click.option("--all", help="Generate gene mappings for every score set in MaveDB.", is_flag=True)
def generate_gene_mappings(db: Session, urns: Sequence[Optional[str]], all: bool):
    score_set_ids: Sequence[Optional[int]]
    if all:
        score_set_ids = db.scalars(select(ScoreSet.id)).all()
        logger.info(
            f"Command invoked with --all. Routine will generate gene mappings for {len(score_set_ids)} score sets."
        )
    else:
        score_set_ids = db.scalars(select(ScoreSet.id).where(ScoreSet.urn.in_(urns))).all()
        logger.info(f"Generating gene mappings for the provided score sets ({len(score_set_ids)}).")

    for ss_id in score_set_ids:
        if not ss_id:
            continue

        score_set = db.scalar(select(ScoreSet).where(ScoreSet.id == ss_id))
        if not score_set:
            logger.warning(f"Could not fetch score set with id={ss_id}.")
            continue

        try:
            mapped_variant = db.scalars(
                select(MappedVariant)
                .join(Variant)
                .where(
                    Variant.score_set_id == ss_id,
                    MappedVariant.current.is_(True),
                    MappedVariant.post_mapped.isnot(None),
                )
                .limit(1)
            ).one_or_none()

            if not mapped_variant:
                logger.info(f"No current mapped variant found for score set {score_set.urn}.")
                continue

            # From line 69, this object must not be None.
            hgvs_string = mapped_variant.post_mapped.get("expressions", {})[0].get("value")  # type: ignore
            if not hgvs_string:
                logger.warning(f"No HGVS string found in post_mapped for variant {mapped_variant.id}.")
                continue

            gene_symbol = get_gene_symbol_from_clingen(hgvs_string)
            if not gene_symbol:
                logger.warning(f"No gene symbol found for HGVS string {hgvs_string}.")
                continue

            # This script has been designed to work prior to the introduction of multi-target mapping.
            # This .one() call reflects those assumptions.
            target_gene = db.scalars(select(TargetGene).where(TargetGene.score_set_id == ss_id)).one()

            if target_gene.post_mapped_metadata is None:
                logger.warning(
                    f"Target gene for score set {score_set.urn} has no post_mapped_metadata despite containing current mapped variants."
                )
                continue

            # Cannot update JSONB fields directly. They can be converted to dictionaries over Mypy's objections.
            post_mapped_metadata = json.loads(json.dumps(dict(target_gene.post_mapped_metadata.copy())))  # type: ignore
            if "genomic" in post_mapped_metadata:
                key = "genomic"
            elif "protein" in post_mapped_metadata:
                key = "protein"
            else:
                logger.warning(f"Unknown post_mapped type for variant {mapped_variant.id}.")

            if "sequence_genes" not in post_mapped_metadata[key]:
                post_mapped_metadata[key]["sequence_genes"] = []
            post_mapped_metadata[key]["sequence_genes"].append(gene_symbol)

            target_gene.post_mapped_metadata = post_mapped_metadata

            db.add(target_gene)
            db.commit()
            logger.info(f"Gene symbol {gene_symbol} added to target gene for score set {score_set.urn}.")

        except Exception as e:
            logger.error(f"Failed to generate gene mappings for score set {score_set.urn}: {str(e)}")
            db.rollback()

    logger.info("Done generating gene mappings.")


if __name__ == "__main__":
    generate_gene_mappings()



import logging as module_logging

import mavedb.logging as application_logging

application_logging.configure()
logger = module_logging.getLogger(__name__)

__project__ = "mavedb-api"
__version__ = "2025.1.1"

logger.info(f"MaveDB {__version__}")



from datetime import date

from sqlalchemy import Boolean, Column, Date, ForeignKey, Integer, String
from sqlalchemy.orm import Mapped, relationship

from mavedb.db.base import Base

from .genome_identifier import GenomeIdentifier


class Taxonomy(Base):
    __tablename__ = "taxonomies"

    id = Column(Integer, primary_key=True)
    tax_id = Column(Integer, nullable=False)
    organism_name = Column(String, nullable=True)
    common_name = Column(String, nullable=True)
    rank = Column(String, nullable=True)
    has_described_species_name = Column(Boolean, nullable=True)
    url = Column(String, nullable=False)
    article_reference = Column(String, nullable=True)
    genome_identifier_id = Column(Integer, ForeignKey("genome_identifiers.id"), index=True, nullable=True)
    genome_identifier: Mapped[GenomeIdentifier] = relationship("GenomeIdentifier", backref="taxonomy_genomes")
    creation_date = Column(Date, nullable=False, default=date.today)
    modification_date = Column(Date, nullable=False, default=date.today, onupdate=date.today)



from typing import Optional

from mavedb.view_models import record_type_validator, set_record_type
from mavedb.view_models.base.base import BaseModel


class OrcidAuthTokenRequest(BaseModel):
    code: str
    redirect_uri: str


class OrcidAuthTokenResponse(BaseModel):
    access_token: str
    expires_in: int
    id_token: str
    token_type: str


class OrcidUser(BaseModel):
    record_type: str = None  # type: ignore
    orcid_id: str
    given_name: Optional[str]
    family_name: Optional[str]

    _record_type_factory = record_type_validator()(set_record_type)



import logging

from sqlalchemy.orm import Session

from mavedb.lib import orcid
from mavedb.lib.exceptions import NonexistentOrcidUserError
from mavedb.lib.logging.context import logging_context, save_to_logging_context
from mavedb.models.contributor import Contributor

logger = logging.getLogger(__name__)


async def find_or_create_contributor(db: Session, orcid_id: str):
    """
    Find an existing contributor record with the specified ORCID ID, or create a new one and fetch the contributor name
    from ORCID.

    :param db: An active database session
    :param orcid_id: A valid ORCID ID
    :return: An existing Contributor with the specified ORCID ID, or a new, unsaved Contributor
    """
    save_to_logging_context({"requested_orcid_user": orcid_id})
    contributor = db.query(Contributor).filter(Contributor.orcid_id == orcid_id).one_or_none()
    if contributor is None:
        # Don't import fetch_orcid_user above, because we want to be able to patch it for unit testing.
        orcid_user = orcid.fetch_orcid_user(orcid_id)
        if orcid_user is None:
            logger.debug(msg="No ORCID user was found for the requested ORCID ID.", extra=logging_context())
            raise NonexistentOrcidUserError(f"No ORCID user was found for ORCID ID {orcid_id}.")
        contributor = Contributor(
            orcid_id=orcid_id, family_name=orcid_user.family_name, given_name=orcid_user.given_name
        )
    return contributor



import logging
import os
import sys
import time
import traceback
from typing import Any, Optional, Union

from starlette.requests import HTTPConnection, Request
from starlette_context import context
from starlette_context.middleware import RawContextMiddleware

from mavedb import __project__, __version__
from mavedb.lib.logging.models import Source

FRONTEND_URL = os.getenv("FRONTEND_URL", "")
API_URL = os.getenv("API_URL", "")

logger = logging.getLogger(__name__)


class PopulatedRawContextMiddleware(RawContextMiddleware):
    async def set_context(self, request: Union[Request, HTTPConnection]) -> dict:
        ctx: dict[str, Any] = {}

        ctx["request_ns"] = time.time_ns()
        ctx["path"] = request.url.path

        if isinstance(request, Request):
            ctx["method"] = request.method
        else:
            try:
                ctx["method"] = request.scope["method"]
            except KeyError:
                pass

        source = Source.other
        if request.headers.get("origin") == FRONTEND_URL:
            source = Source.web
        elif request.headers.get("referer") == API_URL + "/docs":
            source = Source.docs

        ctx["source"] = source
        ctx["application"] = __project__
        ctx["version"] = __version__

        # Retain plugin functionality.
        plugin_ctx = {plugin.key: await plugin.process_request(request) for plugin in self.plugins}

        return {**ctx, **plugin_ctx}


def save_to_logging_context(ctx: dict) -> dict:
    if not context.exists():
        logger.debug("Skipped saving to context. Context does not exist.")
        return {}

    for k, v in ctx.items():
        # Don't overwrite existing context mappings but create a list if a duplicated key is added.
        if k in context:
            existing_ctx = context[k]
            if isinstance(existing_ctx, list):
                context[k].append(v)
            else:
                context[k] = [existing_ctx, v]
        else:
            context[k] = v

    return context.data


def logging_context() -> dict:
    if not context.exists():
        logger.debug("Could not access logging context. Context does not exist.")
        return {}

    return context.data


def correlation_id_for_context() -> Optional[str]:
    return logging_context().get("X-Correlation-ID", None)


def format_raised_exception_info_as_dict(err: BaseException) -> dict:
    _, _, tb = sys.exc_info()

    exc_ctx: dict = {
        "captured_exception_info": {
            "type": err.__class__.__name__,
            "string": str(err),
        }
    }

    try:
        exc_ctx["captured_exception_info"] = {
            **exc_ctx["captured_exception_info"],
            **[
                {"file": fs.filename, "line": fs.lineno, "func": fs.name}
                for fs in traceback.extract_tb(tb)
                # attempt to show only *our* code, not the many layers of library code
                if "/mavedb/" in fs.filename and "/.direnv/" not in fs.filename
            ][-1],
        }

    # We did our best to construct useful traceback info
    except IndexError:
        pass

    return exc_ctx



# Prevent circular imports
from typing import TYPE_CHECKING

from sqlalchemy import Enum, Column, ForeignKey, Integer
from sqlalchemy.orm import Mapped, relationship

from mavedb.db.base import Base
from mavedb.models.enums.contribution_role import ContributionRole

if TYPE_CHECKING:
    from mavedb.models.user import User
    from mavedb.models.collection import Collection


class CollectionUserAssociation(Base):
    __tablename__ = "collection_user_associations"

    collection_id = Column("collection_id", Integer, ForeignKey("collections.id"), primary_key=True)
    user_id = Column("user_id", Integer, ForeignKey("users.id"), primary_key=True)
    contribution_role: Mapped["ContributionRole"] = Column(
        Enum(
            ContributionRole,
            create_constraint=True,
            length=32,
            native_enum=False,
            validate_strings=True,
        ),
        nullable=False,
    )

    collection: Mapped["Collection"] = relationship(
        "mavedb.models.collection.Collection", back_populates="user_associations"
    )
    user: Mapped["User"] = relationship("User")



import logging
import secrets
from typing import Any

from cryptography.hazmat.backends import default_backend as crypto_default_backend
from cryptography.hazmat.primitives import serialization as crypto_serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from fastapi import APIRouter, Depends
from fastapi.encoders import jsonable_encoder
from fastapi.exceptions import HTTPException
from sqlalchemy.orm import Session

from mavedb import deps
from mavedb.lib.authentication import UserData
from mavedb.lib.authorization import require_current_user
from mavedb.lib.logging import LoggedRoute
from mavedb.lib.logging.context import logging_context, save_to_logging_context
from mavedb.models.access_key import AccessKey
from mavedb.models.enums.user_role import UserRole
from mavedb.view_models import access_key

router = APIRouter(
    prefix="/api/v1",
    tags=["access keys"],
    responses={404: {"description": "Not found"}},
    route_class=LoggedRoute,
)

logger = logging.getLogger(__name__)


def generate_key_pair():
    key = rsa.generate_private_key(backend=crypto_default_backend(), public_exponent=65537, key_size=2048)

    private_key = key.private_bytes(
        crypto_serialization.Encoding.PEM,
        crypto_serialization.PrivateFormat.PKCS8,
        crypto_serialization.NoEncryption(),
    )

    public_key = key.public_key().public_bytes(
        crypto_serialization.Encoding.OpenSSH, crypto_serialization.PublicFormat.OpenSSH
    )

    return private_key, public_key


@router.get(
    "/users/me/access-keys",
    status_code=200,
    response_model=list[access_key.AccessKey],
    responses={404: {}, 500: {}},
)
def list_my_access_keys(*, user_data: UserData = Depends(require_current_user)) -> Any:
    """
    List the current user's access keys.
    """
    return user_data.user.access_keys


@router.post(
    "/users/me/access-keys",
    status_code=200,
    response_model=access_key.NewAccessKey,
    responses={404: {}, 500: {}},
)
def create_my_access_key(
    *,
    db: Session = Depends(deps.get_db),
    user_data: UserData = Depends(require_current_user),
) -> Any:
    """
    Create a new access key for the current user, with the default user role.
    """
    private_key, public_key = generate_key_pair()

    item = AccessKey(user=user_data.user, key_id=secrets.token_urlsafe(32), public_key=public_key)
    db.add(item)
    db.commit()
    db.refresh(item)

    response_item = access_key.NewAccessKey(**jsonable_encoder(item, by_alias=False), private_key=private_key)
    response_item.private_key = private_key
    return response_item


@router.post(
    "/users/me/access-keys/{role}",
    status_code=200,
    response_model=access_key.NewAccessKey,
    responses={404: {}, 500: {}},
)
async def create_my_access_key_with_role(
    *,
    role: UserRole,
    db: Session = Depends(deps.get_db),
    user_data: UserData = Depends(require_current_user),
) -> Any:
    """
    Create a new access key for the current user, with the specified role.
    """
    save_to_logging_context({"requested_role": role.name})
    # Allow the user to create an access key for any of their potential roles, not just their active one.
    if not any(user_role == role for user_role in user_data.user.roles):
        logger.warning(
            msg="Could not create API key for user; User does not belong to the requested role.",
            extra=logging_context(),
        )

        raise HTTPException(
            status_code=403,
            detail="User cannot create an API key for a role they do not have.",
        )

    private_key, public_key = generate_key_pair()

    item = AccessKey(user=user_data.user, key_id=secrets.token_urlsafe(32), public_key=public_key)
    await item.set_role(db, role)
    db.add(item)
    db.commit()
    db.refresh(item)

    response_item = access_key.NewAccessKey(**jsonable_encoder(item, by_alias=False), private_key=private_key)
    response_item.private_key = private_key
    return response_item


@router.delete("/users/me/access-keys/{key_id}", status_code=200, responses={404: {}, 500: {}})
def delete_my_access_key(
    *,
    key_id: str,
    db: Session = Depends(deps.get_db),
    user_data: UserData = Depends(require_current_user),
) -> Any:
    """
    Delete one of the current user's access keys.
    """
    item = db.query(AccessKey).filter(AccessKey.key_id == key_id).one_or_none()
    if item and item.user.id == user_data.user.id:
        db.delete(item)
        db.commit()
        logger.debug(msg="Successfully deleted provided API key.", extra=logging_context())



# Prevent circular imports
from typing import TYPE_CHECKING

from sqlalchemy import Boolean, Column, ForeignKey, Integer
from sqlalchemy.orm import Mapped, relationship

from mavedb.db.base import Base

if TYPE_CHECKING:
    from mavedb.models.publication_identifier import PublicationIdentifier
    from mavedb.models.score_set import ScoreSet


class ScoreSetPublicationIdentifierAssociation(Base):
    __tablename__ = "scoreset_publication_identifiers"

    score_set_id = Column("scoreset_id", Integer, ForeignKey("scoresets.id"), primary_key=True)
    publication_identifier_id = Column(Integer, ForeignKey("publication_identifiers.id"), primary_key=True)
    primary = Column(Boolean, nullable=True, default=False)

    score_set: Mapped["ScoreSet"] = relationship(
        "mavedb.models.score_set.ScoreSet", back_populates="publication_identifier_associations"
    )
    publication: Mapped["PublicationIdentifier"] = relationship("PublicationIdentifier")



import itertools
from collections import OrderedDict, Counter
from enum import Enum
from typing import Any, Union, Optional

from fastapi import APIRouter, Depends, HTTPException
from sqlalchemy import Table, func, select, Select
from sqlalchemy.orm import Session

from mavedb.deps import get_db
from mavedb.models.controlled_keyword import ControlledKeyword
from mavedb.models.doi_identifier import DoiIdentifier
from mavedb.models.ensembl_identifier import EnsemblIdentifier
from mavedb.models.ensembl_offset import EnsemblOffset
from mavedb.models.experiment import (
    Experiment,
    experiments_doi_identifiers_association_table,
    experiments_raw_read_identifiers_association_table,
)
from mavedb.models.experiment_controlled_keyword import ExperimentControlledKeywordAssociation
from mavedb.models.experiment_publication_identifier import ExperimentPublicationIdentifierAssociation
from mavedb.models.publication_identifier import PublicationIdentifier
from mavedb.models.published_variant import PublishedVariantsMV
from mavedb.models.raw_read_identifier import RawReadIdentifier
from mavedb.models.refseq_identifier import RefseqIdentifier
from mavedb.models.refseq_offset import RefseqOffset
from mavedb.models.score_set import (
    ScoreSet,
    score_sets_doi_identifiers_association_table,
    score_sets_raw_read_identifiers_association_table,
)
from mavedb.models.score_set_publication_identifier import ScoreSetPublicationIdentifierAssociation
from mavedb.models.target_accession import TargetAccession
from mavedb.models.target_gene import TargetGene
from mavedb.models.target_sequence import TargetSequence
from mavedb.models.taxonomy import Taxonomy
from mavedb.models.uniprot_identifier import UniprotIdentifier
from mavedb.models.uniprot_offset import UniprotOffset
from mavedb.models.user import User

router = APIRouter(
    prefix="/api/v1/statistics",
    tags=["statistics"],
    responses={404: {"description": "Not found"}},
)

TARGET_ACCESSION_TAXONOMY = "Homo sapiens"

## Union types

RecordModels = Union[type[Experiment], type[ScoreSet]]
RecordAssociationTables = Union[
    Table,
    type[ExperimentControlledKeywordAssociation],
    type[ExperimentPublicationIdentifierAssociation],
    type[ScoreSetPublicationIdentifierAssociation],
]


## Enum classes hold valid endpoints for different statistics routes.


class RecordNames(str, Enum):
    experiment = "experiment"
    scoreSet = "score-set"


class RecordFields(str, Enum):
    publicationIdentifiers = "publication-identifiers"
    keywords = "keywords"
    doiIdentifiers = "doi-identifiers"
    rawReadIdentifiers = "raw-read-identifiers"
    createdBy = "created-by"


class GroupBy(str, Enum):
    month = "month"
    year = "year"


def _model_and_association_from_record_field(
    record: RecordNames, field: Optional[RecordFields]
) -> tuple[RecordModels, Optional[RecordAssociationTables]]:
    """
    Given a member of the RecordNames and RecordFields Enums, generate the model and association table that can be used
    to generate statistics for those fields.

    This function should be used for generating statistics for fields shared between Experiments and Score Sets.
    If necessary, Experiment Sets can be handled in a similar manner in the future.
    """
    record_to_model_map: dict[RecordNames, RecordModels] = {
        RecordNames.experiment: Experiment,
        RecordNames.scoreSet: ScoreSet,
    }
    record_to_assc_map: dict[RecordNames, dict[RecordFields, RecordAssociationTables]] = {
        RecordNames.experiment: {
            RecordFields.doiIdentifiers: experiments_doi_identifiers_association_table,
            RecordFields.publicationIdentifiers: ExperimentPublicationIdentifierAssociation,
            RecordFields.rawReadIdentifiers: experiments_raw_read_identifiers_association_table,
            RecordFields.keywords: ExperimentControlledKeywordAssociation,
        },
        RecordNames.scoreSet: {
            RecordFields.doiIdentifiers: score_sets_doi_identifiers_association_table,
            RecordFields.publicationIdentifiers: ScoreSetPublicationIdentifierAssociation,
            RecordFields.rawReadIdentifiers: score_sets_raw_read_identifiers_association_table,
        },
    }

    queried_model = record_to_model_map[record]
    queried_model_assc = record_to_assc_map[record]

    if field is None or field not in queried_model_assc:
        return queried_model, None

    return queried_model, queried_model_assc[field]


def _join_model_and_filter_unpublished(query: Select, model: RecordModels) -> Select:
    return query.join(model).where(model.published_date.is_not(None))


def _count_for_identifier_in_query(db: Session, query: Select[tuple[Any, int]]) -> dict[Any, int]:
    return {value: count for value, count in db.execute(query).all() if value is not None}


########################################################################################
#  Record statistics
########################################################################################


@router.get(
    "/record/{record}/keywords", status_code=200, response_model=Union[dict[str, int], dict[str, dict[str, int]]]
)
def experiment_keyword_statistics(
    record: RecordNames, db: Session = Depends(get_db)
) -> Union[dict[str, int], dict[str, dict[str, int]]]:
    """
    Returns a dictionary of counts for the distinct values of the `value` field (member of the `controlled_keywords` table).
    Don't include any NULL field values. Don't include any keywords from unpublished experiments.
    """
    if record == RecordNames.scoreSet:
        raise HTTPException(
            422,
            "The 'keywords' field can only be used with the 'experiment' model. Score sets do not have associated keywords.",
        )

    queried_model, queried_assc = _model_and_association_from_record_field(record, RecordFields.keywords)

    if queried_assc is None:
        raise HTTPException(500, "No association table associated with the keywords field when one was expected.")

    query = _join_model_and_filter_unpublished(
        select(ControlledKeyword.value, func.count(ControlledKeyword.value)).join(queried_assc), queried_model
    ).group_by(ControlledKeyword.value)

    return _count_for_identifier_in_query(db, query)


@router.get("/record/{record}/publication-identifiers", status_code=200, response_model=dict[str, dict[str, int]])
def experiment_publication_identifier_statistics(
    record: RecordNames, db: Session = Depends(get_db)
) -> dict[str, dict[str, int]]:
    """
    Returns a dictionary of counts for the distinct values of the `identifier` field (member of the `publication_identifiers` table).
    Don't include any publication identifiers from unpublished experiments.
    """
    queried_model, queried_assc = _model_and_association_from_record_field(record, RecordFields.publicationIdentifiers)

    if queried_assc is None:
        raise HTTPException(
            500, "No association table associated with the publication identifiers field when one was expected."
        )

    query = _join_model_and_filter_unpublished(
        select(
            PublicationIdentifier.identifier,
            PublicationIdentifier.db_name,
            func.count(PublicationIdentifier.identifier),
        ).join(queried_assc),
        queried_model,
    ).group_by(PublicationIdentifier.identifier, PublicationIdentifier.db_name)

    publication_identifiers: dict[str, dict[str, int]] = {}

    for identifier, db_name, count in db.execute(query).all():
        # We don't need to worry about overwriting existing identifiers within these internal dictionaries because
        # of the SQL group by clause.
        if db_name in publication_identifiers:
            publication_identifiers[db_name][identifier] = count
        else:
            publication_identifiers[db_name] = {identifier: count}

    return publication_identifiers


@router.get("/record/{record}/raw-read-identifiers", status_code=200, response_model=dict[str, int])
def experiment_raw_read_identifier_statistics(record: RecordNames, db: Session = Depends(get_db)) -> dict[str, int]:
    """
    Returns a dictionary of counts for the distinct values of the `identifier` field (member of the `raw_read_identifiers` table).
    Don't include any raw read identifiers from unpublished experiments.
    """
    queried_model, queried_assc = _model_and_association_from_record_field(record, RecordFields.rawReadIdentifiers)

    if queried_assc is None:
        raise HTTPException(
            500, "No association table associated with the raw read identifiers field when one was expected."
        )

    query = _join_model_and_filter_unpublished(
        select(RawReadIdentifier.identifier, func.count(RawReadIdentifier.identifier)).join(queried_assc), queried_model
    ).group_by(RawReadIdentifier.identifier)

    return _count_for_identifier_in_query(db, query)


@router.get("/record/{record}/doi-identifiers", status_code=200, response_model=dict[str, int])
def experiment_doi_identifiers_statistics(record: RecordNames, db: Session = Depends(get_db)) -> dict[str, int]:
    """
    Returns a dictionary of counts for the distinct values of the `identifier` field (member of the `doi_identifiers` table).
    Don't include any DOI identifiers from unpublished experiments.
    """
    queried_model, queried_assc = _model_and_association_from_record_field(record, RecordFields.doiIdentifiers)

    if queried_assc is None:
        raise HTTPException(
            500, "No association table associated with the doi identifiers field when one was expected."
        )

    query = _join_model_and_filter_unpublished(
        select(DoiIdentifier.identifier, func.count(DoiIdentifier.identifier)).join(queried_assc), queried_model
    ).group_by(DoiIdentifier.identifier)

    return _count_for_identifier_in_query(db, query)


@router.get("/record/{record}/created-by", status_code=200, response_model=dict[str, int])
def experiment_created_by_statistics(record: RecordNames, db: Session = Depends(get_db)) -> dict[str, int]:
    """
    Returns a dictionary of counts for the distinct values of the `username` field (member of the `users` table).
    Don't include any usernames from unpublished experiments.
    """
    queried_model, queried_assc = _model_and_association_from_record_field(record, RecordFields.createdBy)

    query = (
        select(User.username, func.count(User.id))
        .join(queried_model, queried_model.created_by_id == User.id)
        .filter(queried_model.published_date.is_not(None))
        .group_by(User.id)
    )

    return _count_for_identifier_in_query(db, query)


@router.get("/record/{model}/published/count", status_code=200, response_model=dict[str, int])
def record_counts(model: RecordNames, group: Optional[GroupBy] = None, db: Session = Depends(get_db)) -> dict[str, int]:
    """
    Returns a dictionary of counts for the number of published records of the `model` parameter.
    Optionally, group the counts by the published month or year.
    """
    queried_model, queried_assc = _model_and_association_from_record_field(model, None)

    # Protect against Nonetype publication dates with where clause.
    # We can safely ignore Mypy Nonetype errors in the following dictcomps.
    objs = db.scalars(
        select(queried_model.published_date)
        .where(queried_model.published_date.isnot(None))
        .order_by(queried_model.published_date)
    ).all()

    if group == GroupBy.month:
        grouped = {k: len(list(g)) for k, g in itertools.groupby(objs, lambda t: t.strftime("%Y-%m"))}  # type: ignore
    elif group == GroupBy.year:
        grouped = {k: len(list(g)) for k, g in itertools.groupby(objs, lambda t: t.strftime("%Y"))}  # type: ignore
    else:
        grouped = {"count": len(objs)}

    return OrderedDict(sorted(grouped.items()))


@router.get("/record/score-set/variant/count", status_code=200, response_model=dict[str, int])
def record_variant_counts(db: Session = Depends(get_db)) -> dict[str, int]:
    """
    Returns a dictionary of counts for the number of published and distinct variants in the database contained
    within a given record.
    """
    variants = db.execute(
        select(PublishedVariantsMV.score_set_urn, func.count(PublishedVariantsMV.variant_id))
        .group_by(PublishedVariantsMV.score_set_urn)
        .order_by(PublishedVariantsMV.score_set_urn)
    ).all()

    grouped = {urn: sum(c for _, c in g) for urn, g in itertools.groupby(variants, lambda t: t[0])}
    return OrderedDict(sorted(filter(lambda item: item[1] > 0, grouped.items())))


@router.get("/record/score-set/mapped-variant/count", status_code=200, response_model=dict[str, int])
def record_mapped_variant_counts(db: Session = Depends(get_db)) -> dict[str, int]:
    """
    Returns a dictionary of counts for the number of published and distinct mapped variants in the database contained
    within a given record.
    """
    variants = db.execute(
        select(PublishedVariantsMV.score_set_urn, func.count(PublishedVariantsMV.mapped_variant_id))
        .group_by(PublishedVariantsMV.score_set_urn)
        .order_by(PublishedVariantsMV.score_set_urn)
    ).all()

    grouped = {urn: sum(c for _, c in g) for urn, g in itertools.groupby(variants, lambda t: t[0])}
    return OrderedDict(sorted(filter(lambda item: item[1] > 0, grouped.items())))


########################################################################################
# Target statistics
########################################################################################


##### Accession based targets #####


@router.get("/target/accession/accession", status_code=200, response_model=dict[str, int])
def target_accessions_accession_counts(db: Session = Depends(get_db)) -> dict[str, int]:
    """
    Returns a dictionary of counts for the distinct values of the `accession` field (member of the `target_accessions` table).
    Don't include any NULL field values. Don't include any targets from unpublished score sets.
    """
    query = _join_model_and_filter_unpublished(
        select(TargetAccession.accession, func.count(TargetAccession.accession)).join(TargetGene), ScoreSet
    ).group_by(TargetAccession.accession)

    return _count_for_identifier_in_query(db, query)


@router.get("/target/accession/assembly", status_code=200, response_model=dict[str, int])
def target_accessions_assembly_counts(db: Session = Depends(get_db)) -> dict[str, int]:
    """
    Returns a dictionary of counts for the distinct values of the `assembly` field (member of the `target_accessions` table).
    Don't include any NULL field values. Don't include any targets from unpublished score sets.
    """
    query = _join_model_and_filter_unpublished(
        select(TargetAccession.assembly, func.count(TargetAccession.assembly)).join(TargetGene), ScoreSet
    ).group_by(TargetAccession.assembly)

    return _count_for_identifier_in_query(db, query)


@router.get("/target/accession/gene", status_code=200, response_model=dict[str, int])
def target_accessions_gene_counts(db: Session = Depends(get_db)) -> dict[str, int]:
    """
    Returns a dictionary of counts for the distinct values of the `gene` field (member of the `target_accessions` table).
    Don't include any NULL field values. Don't include any targets from unpublished score sets.
    """
    query = _join_model_and_filter_unpublished(
        select(TargetAccession.gene, func.count(TargetAccession.gene)).join(TargetGene), ScoreSet
    ).group_by(TargetAccession.gene)

    return _count_for_identifier_in_query(db, query)


##### Sequence based targets #####


@router.get("/target/sequence/sequence", status_code=200, response_model=dict[str, int])
def target_sequences_sequence_counts(db: Session = Depends(get_db)) -> dict[str, int]:
    """
    Returns a dictionary of counts for the distinct values of the `sequence` field (member of the `target_sequences` table).
    Don't include any NULL field values. Don't include any targets from unpublished score sets.
    """
    query = _join_model_and_filter_unpublished(
        select(TargetSequence.sequence, func.count(TargetSequence.sequence)).join(TargetGene), ScoreSet
    ).group_by(TargetSequence.sequence)

    return _count_for_identifier_in_query(db, query)


@router.get("/target/sequence/sequence-type", status_code=200, response_model=dict[str, int])
def target_sequences_sequence_type_counts(db: Session = Depends(get_db)) -> dict[str, int]:
    """
    Returns a dictionary of counts for the distinct values of the `sequence_type` field (member of the `target_sequences` table).
    Don't include any NULL field values. Don't include any targets from unpublished score sets.
    """
    query = _join_model_and_filter_unpublished(
        select(TargetSequence.sequence_type, func.count(TargetSequence.sequence_type)).join(TargetGene), ScoreSet
    ).group_by(TargetSequence.sequence_type)

    return _count_for_identifier_in_query(db, query)


##### Target genes #####


@router.get("/target/gene/category", status_code=200, response_model=dict[str, int])
def target_genes_category_counts(db: Session = Depends(get_db)) -> dict[str, int]:
    """
    Returns a dictionary of counts for the distinct values of the `category` field (member of the `target_sequences` table).
    Don't include any NULL field values. Don't include any targets from unpublished score sets.
    """
    query = _join_model_and_filter_unpublished(
        select(TargetGene.category, func.count(TargetGene.category)), ScoreSet
    ).group_by(TargetGene.category)

    return _count_for_identifier_in_query(db, query)


@router.get("/target/gene/organism", status_code=200, response_model=dict[str, int])
def target_genes_organism_counts(db: Session = Depends(get_db)) -> dict[str, int]:
    """
    Returns a dictionary of counts for the distinct values of the `organism` field (member of the `taxonomies` table).
    Don't include any NULL field values. Don't include any targets from unpublished score sets.

    NOTE: For now (and perhaps forever), all accession based targets are human genomic sequences (ie: of taxonomy `Homo sapiens`).
          It is possible this assumption changes if we add mouse (or other non-human) genomes to MaveDB.
    """
    target_sequence_query = _join_model_and_filter_unpublished(
        select(Taxonomy.organism_name, func.count(Taxonomy.organism_name)).join(TargetSequence).join(TargetGene),
        ScoreSet,
    ).group_by(Taxonomy.organism_name)
    target_accession_query = _join_model_and_filter_unpublished(
        select(func.count(TargetAccession.id)).join(TargetGene), ScoreSet
    )

    # Ensure the `Homo sapiens` key exists in the organisms counts dictionary.
    organisms = _count_for_identifier_in_query(db, target_sequence_query)
    organisms.setdefault(TARGET_ACCESSION_TAXONOMY, 0)

    count_accession_based_targets = db.execute(target_accession_query).scalar_one_or_none()
    if count_accession_based_targets:
        organisms[TARGET_ACCESSION_TAXONOMY] += count_accession_based_targets
    else:
        organisms.pop(TARGET_ACCESSION_TAXONOMY)

    return organisms


@router.get("/target/gene/ensembl-identifier", status_code=200, response_model=dict[str, int])
def target_genes_ensembl_identifier_counts(db: Session = Depends(get_db)) -> dict[str, int]:
    """
    Returns a dictionary of counts for the distinct values of the `identifier` field (member of the `ensembl_identifiers` table).
    Don't include any NULL field values. Don't include any targets from unpublished score sets.
    """
    query = _join_model_and_filter_unpublished(
        select(EnsemblIdentifier.identifier, func.count(EnsemblIdentifier.identifier))
        .join(EnsemblOffset)
        .join(TargetGene),
        ScoreSet,
    ).group_by(EnsemblIdentifier.identifier)

    return _count_for_identifier_in_query(db, query)


@router.get("/target/gene/refseq-identifier", status_code=200, response_model=dict[str, int])
def target_genes_refseq_identifier_counts(db: Session = Depends(get_db)) -> dict[str, int]:
    """
    Returns a dictionary of counts for the distinct values of the `identifier` field (member of the `refseq_identifiers` table).
    Don't include any NULL field values. Don't include any targets from unpublished score sets.
    """
    query = _join_model_and_filter_unpublished(
        select(RefseqIdentifier.identifier, func.count(RefseqIdentifier.identifier))
        .join(RefseqOffset)
        .join(TargetGene),
        ScoreSet,
    ).group_by(RefseqIdentifier.identifier)

    return _count_for_identifier_in_query(db, query)


@router.get("/target/gene/uniprot-identifier", status_code=200, response_model=dict[str, int])
def target_genes_uniprot_identifier_counts(db: Session = Depends(get_db)) -> dict[str, int]:
    """
    Returns a dictionary of counts for the distinct values of the `identifier` field (member of the `uniprot_identifiers` table).
    Don't include any NULL field values. Don't include any targets from unpublished score sets.
    """
    query = _join_model_and_filter_unpublished(
        select(UniprotIdentifier.identifier, func.count(UniprotIdentifier.identifier))
        .join(UniprotOffset)
        .join(TargetGene),
        ScoreSet,
    ).group_by(UniprotIdentifier.identifier)

    return _count_for_identifier_in_query(db, query)


@router.get("/target/mapped/gene")
def mapped_target_gene_counts(db: Session = Depends(get_db)) -> dict[str, int]:
    """
    Returns a dictionary of counts for the distinct values of the `gene` property within the `post_mapped_metadata`
    field (member of the `target_gene` table). Don't include any NULL field values. Don't include any targets from
    unpublished score sets.
    """
    query = _join_model_and_filter_unpublished(
        select(TargetGene.post_mapped_metadata),
        ScoreSet,
    ).where(TargetGene.post_mapped_metadata.isnot(None))

    mapping_metadata = db.scalars(query).all()
    gene_counts = Counter(
        gene
        for metadata in mapping_metadata
        for key in ("genomic", "protein")
        if key in metadata
        for gene in metadata[key].get("sequence_genes", [])
    )

    # The gene will always be a string
    return dict(gene_counts)  # type: ignore


########################################################################################
# Variant (and mapped variant) statistics
########################################################################################


@router.get("/variant/count", status_code=200, response_model=dict[str, int])
def variant_counts(group: Optional[GroupBy] = None, db: Session = Depends(get_db)) -> dict[str, int]:
    """
    Returns a dictionary of counts for the number of published and distinct variants in the database.
    Optionally, group the counts by the day on which the score set (and by extension, the variant) was published.
    """
    variants = db.execute(
        select(PublishedVariantsMV.published_date, func.count(PublishedVariantsMV.variant_id))
        .group_by(PublishedVariantsMV.published_date)
        .order_by(PublishedVariantsMV.published_date)
    ).all()

    if group == GroupBy.month:
        grouped = {k: sum(c for _, c in g) for k, g in itertools.groupby(variants, lambda t: t[0].strftime("%Y-%m"))}
    elif group == GroupBy.year:
        grouped = {k: sum(c for _, c in g) for k, g in itertools.groupby(variants, lambda t: t[0].strftime("%Y"))}
    else:
        grouped = {"count": sum(count for _, count in variants)}

    return OrderedDict(sorted(grouped.items()))


@router.get("/mapped-variant/count", status_code=200, response_model=dict[str, int])
def mapped_variant_counts(
    group: Optional[GroupBy] = None, onlyCurrent: bool = True, db: Session = Depends(get_db)
) -> dict[str, int]:
    """
    Returns a dictionary of counts for the number of published and distinct variants in the database.
    Optionally, group the counts by the day on which the score set (and by extension, the variant) was published.
    Optionally, return the count of all mapped variants, not just the current/most up to date ones.
    """
    query = select(PublishedVariantsMV.published_date, func.count(PublishedVariantsMV.mapped_variant_id))

    if onlyCurrent:
        query = query.where(PublishedVariantsMV.current_mapped_variant.is_(True))

    variants = db.execute(
        query.group_by(PublishedVariantsMV.published_date).order_by(PublishedVariantsMV.published_date)
    ).all()

    if group == GroupBy.month:
        grouped = {k: sum(c for _, c in g) for k, g in itertools.groupby(variants, lambda t: t[0].strftime("%Y-%m"))}
    elif group == GroupBy.year:
        grouped = {k: sum(c for _, c in g) for k, g in itertools.groupby(variants, lambda t: t[0].strftime("%Y"))}
    else:
        grouped = {"count": sum(count for _, count in variants)}

    return OrderedDict(sorted(grouped.items()))



from datetime import date
from typing import Any, Optional

from mavedb.view_models import record_type_validator, set_record_type
from mavedb.view_models.base.base import BaseModel


class MappedVariantBase(BaseModel):
    pre_mapped: Optional[Any]
    post_mapped: Optional[Any]
    variant_id: int
    vrs_version: Optional[str]
    error_message: Optional[str]
    modification_date: date
    mapped_date: date
    mapping_api_version: str
    current: bool


class MappedVariantCreate(MappedVariantBase):
    pass


class MappedVariantUpdate(MappedVariantBase):
    pass


# Properties shared by models stored in DB
class SavedMappedVariant(MappedVariantBase):
    id: int
    record_type: str = None  # type: ignore

    _record_type_factory = record_type_validator()(set_record_type)

    class Config:
        orm_mode = True


# Properties to return to non-admin clients
class MappedVariant(SavedMappedVariant):
    pass



from datetime import date

from sqlalchemy import Boolean, Column, Date, ForeignKey, Integer, String
from sqlalchemy.ext.associationproxy import AssociationProxy, association_proxy
from sqlalchemy.orm import Mapped, relationship

import mavedb.models.collection_user_association
from mavedb.db.base import Base
from mavedb.lib.urns import generate_collection_urn
from mavedb.models.collection_association import (
    collection_experiments_association_table,
    collection_score_sets_association_table,
)

from .experiment import Experiment
from .score_set import ScoreSet
from .user import User


class Collection(Base):
    __tablename__ = "collections"

    id = Column(Integer, primary_key=True)

    urn = Column(
        String(64),
        nullable=True,
        default=generate_collection_urn,
        unique=True,
        index=True,
    )
    private = Column(Boolean, nullable=False, default=True)

    name = Column(String, nullable=False)
    badge_name = Column(String, nullable=True)
    description = Column(String, nullable=True)

    created_by_id = Column(Integer, ForeignKey("users.id"), index=True, nullable=True)
    created_by: Mapped[User] = relationship("User", foreign_keys="Collection.created_by_id")
    modified_by_id = Column(Integer, ForeignKey("users.id"), index=True, nullable=True)
    modified_by: Mapped[User] = relationship("User", foreign_keys="Collection.modified_by_id")
    creation_date = Column(Date, nullable=False, default=date.today)
    modification_date = Column(Date, nullable=False, default=date.today, onupdate=date.today)

    user_associations: Mapped[list[mavedb.models.collection_user_association.CollectionUserAssociation]] = relationship(
        "CollectionUserAssociation",
        back_populates="collection",
        cascade="all, delete-orphan",
    )
    users: AssociationProxy[list[User]] = association_proxy(
        "user_associations",
        "user",
        creator=lambda u: mavedb.models.collection_user_association.CollectionUserAssociation(
            user=u, contribution_role=u.role
        ),
    )

    experiments: Mapped[list[Experiment]] = relationship(
        "Experiment",
        secondary=collection_experiments_association_table,
        back_populates="collections",
    )
    score_sets: Mapped[list[ScoreSet]] = relationship(
        "ScoreSet",
        secondary=collection_score_sets_association_table,
        back_populates="collections",
    )



from datetime import date
from typing import Any, Collection, Optional, Sequence

from mavedb.lib.validation.exceptions import ValidationError
from mavedb.lib.validation.utilities import is_null
from mavedb.view_models import PublicationIdentifiersGetter, record_type_validator, set_record_type
from mavedb.view_models.base.base import BaseModel, validator
from mavedb.view_models.contributor import Contributor, ContributorCreate
from mavedb.view_models.doi_identifier import (
    DoiIdentifier,
    DoiIdentifierCreate,
    SavedDoiIdentifier,
)
from mavedb.view_models.experiment_controlled_keyword import (
    ExperimentControlledKeyword,
    ExperimentControlledKeywordCreate,
    SavedExperimentControlledKeyword,
)
from mavedb.view_models.publication_identifier import (
    PublicationIdentifier,
    PublicationIdentifierCreate,
    SavedPublicationIdentifier,
)
from mavedb.view_models.raw_read_identifier import (
    RawReadIdentifier,
    RawReadIdentifierCreate,
    SavedRawReadIdentifier,
)
from mavedb.view_models.user import SavedUser, User


class OfficialCollection(BaseModel):
    badge_name: str
    name: str
    urn: str

    class Config:
        orm_mode = True
        arbitrary_types_allowed = True


class ExperimentGetter(PublicationIdentifiersGetter):
    def get(self, key: Any, default: Any = ...) -> Any:
        if key == "score_set_urns":
            score_sets = getattr(self._obj, "score_sets") or []
            return sorted([score_set.urn for score_set in score_sets if score_set.superseding_score_set is None])
        elif key == "experiment_set_urn":
            experiment_set = getattr(self._obj, "experiment_set")
            return experiment_set.urn if experiment_set is not None else None
        else:
            return super().get(key, default)


class ExperimentBase(BaseModel):
    title: str
    short_description: str
    abstract_text: Optional[str]
    method_text: Optional[str]
    extra_metadata: Optional[dict]

    @classmethod
    def from_orm(cls, obj: Any):
        try:
            obj.experiment_set_urn = obj.experiment_set.urn
        except AttributeError:
            obj.experiment_set_urn = None
        return super().from_orm(obj)


class ExperimentModify(ExperimentBase):
    abstract_text: str
    method_text: str
    contributors: Optional[list[ContributorCreate]]
    keywords: Optional[list[ExperimentControlledKeywordCreate]]
    doi_identifiers: Optional[list[DoiIdentifierCreate]]
    primary_publication_identifiers: Optional[list[PublicationIdentifierCreate]]
    secondary_publication_identifiers: Optional[list[PublicationIdentifierCreate]]
    raw_read_identifiers: Optional[list[RawReadIdentifierCreate]]

    @validator("primary_publication_identifiers")
    def max_one_primary_publication_identifier(cls, v):
        if isinstance(v, list):
            if len(v) > 1:
                raise ValidationError("multiple primary publication identifiers are not allowed")
        return v

    @validator("title", "short_description", "abstract_text", "method_text")
    def validate_field_is_non_empty(cls, v):
        if is_null(v) or not isinstance(v, str):
            raise ValidationError("This field is required and cannot be empty.")
        return v.strip()


class ExperimentCreate(ExperimentModify):
    experiment_set_urn: Optional[str]


class ExperimentUpdate(ExperimentModify):
    pass


# Properties shared by models stored in DB
class SavedExperiment(ExperimentBase):
    record_type: str = None  # type: ignore
    urn: str
    created_by: SavedUser
    modified_by: SavedUser
    creation_date: date
    modification_date: date
    published_date: Optional[date]
    experiment_set_urn: str
    doi_identifiers: Sequence[SavedDoiIdentifier]
    primary_publication_identifiers: Sequence[SavedPublicationIdentifier]
    secondary_publication_identifiers: Sequence[SavedPublicationIdentifier]
    raw_read_identifiers: Sequence[SavedRawReadIdentifier]
    contributors: list[Contributor]
    keywords: Sequence[SavedExperimentControlledKeyword]

    _record_type_factory = record_type_validator()(set_record_type)

    class Config:
        orm_mode = True
        getter_dict = ExperimentGetter

    # Association proxy objects return an untyped _AssocitionList object.
    # Recast it into something more generic.
    @validator("secondary_publication_identifiers", "primary_publication_identifiers", pre=True)
    def publication_identifiers_validator(cls, value, values, field) -> list[PublicationIdentifier]:
        assert isinstance(value, Collection), f"{field.name} must be a collection, not {type(value)}"
        return list(value)  # Re-cast into proper list-like type


# Properties to return to non-admin clients
class Experiment(SavedExperiment):
    score_set_urns: list[str]
    processing_state: Optional[str]
    doi_identifiers: Sequence[DoiIdentifier]
    primary_publication_identifiers: Sequence[PublicationIdentifier]
    secondary_publication_identifiers: Sequence[PublicationIdentifier]
    official_collections: Sequence[OfficialCollection]
    keywords: Sequence[ExperimentControlledKeyword]
    raw_read_identifiers: Sequence[RawReadIdentifier]
    created_by: User
    modified_by: User


class ShortExperiment(SavedExperiment):
    score_set_urns: list[str]
    processing_state: Optional[str]


# Properties to return to admin clients
class AdminExperiment(Experiment):
    score_set_urns: list[str]
    processing_state: Optional[str]
    approved: bool


# Properties to include in a dump of all published data.
class ExperimentPublicDump(SavedExperiment):
    score_sets: "Sequence[ScoreSetPublicDump]"


# ruff: noqa: E402
from mavedb.view_models.score_set import ScoreSetPublicDump

ExperimentPublicDump.update_forward_refs()



import logging
import os
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import Optional

from fastapi import Depends, Header, HTTPException, Request, Security
from fastapi.security import (
    APIKeyCookie,
    APIKeyHeader,
    APIKeyQuery,
    HTTPAuthorizationCredentials,
    HTTPBearer,
)
from jose import jwt
from sqlalchemy.orm import Session

from mavedb import deps
from mavedb.lib.logging.context import format_raised_exception_info_as_dict, logging_context, save_to_logging_context
from mavedb.lib.orcid import fetch_orcid_user_email
from mavedb.models.access_key import AccessKey
from mavedb.models.enums.user_role import UserRole
from mavedb.models.user import User

ORCID_JWT_SIGNING_PUBLIC_KEY = """-----BEGIN PUBLIC KEY-----
MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAjxTIntA7YvdfnYkLSN4w
k//E2zf/wbb0SV/HLHFvh6a9ENVRD1/rHK0EijlBzikb+1rgDQihJETcgBLsMoZV
QqGj8fDUUuxnVHsuGav/bf41PA7E/58HXKPrB2C0cON41f7K3o9TStKpVJOSXBrR
WURmNQ64qnSSryn1nCxMzXpaw7VUo409ohybbvN6ngxVy4QR2NCC7Fr0QVdtapxD
7zdlwx6lEwGemuqs/oG5oDtrRuRgeOHmRps2R6gG5oc+JqVMrVRv6F9h4ja3UgxC
DBQjOVT1BFPWmMHnHCsVYLqbbXkZUfvP2sO1dJiYd/zrQhi+FtNth9qrLLv3gkgt
wQIDAQAB
-----END PUBLIC KEY-----
"""
ORCID_JWT_AUDIENCE = os.getenv("ORCID_CLIENT_ID")

ACCESS_TOKEN_NAME = "X-API-key"

logger = logging.getLogger(__name__)


class AuthenticationMethod(str, Enum):
    api_key = "api_key"
    jwt = "jwt"


@dataclass
class UserData:
    user: User
    active_roles: list[UserRole]


####################################################################################################
# JWT authentication
####################################################################################################


def decode_jwt(token: str) -> dict:
    try:
        decoded_token = jwt.decode(
            token,
            ORCID_JWT_SIGNING_PUBLIC_KEY,
            algorithms=["RS256"],
            audience=ORCID_JWT_AUDIENCE,
            # ORCID sends an at_hash when using the OpenID Connect implicit flow, even though there is no auth_token.
            options={"verify_at_hash": False},
        )
        return decoded_token
    # TODO: should catch specific exceptions, and should log them more usefully.
    except Exception as ex:
        save_to_logging_context(format_raised_exception_info_as_dict(ex))
        logger.debug(msg="Failed to authenticate user; Could not decode user token.", extra=logging_context())
        return {}


class JWTBearer(HTTPBearer):
    def __init__(self, auto_error: bool = True):
        super(JWTBearer, self).__init__(auto_error=auto_error)

    async def __call__(self, request: Request):
        credentials: Optional[HTTPAuthorizationCredentials]
        try:
            credentials = await super(JWTBearer, self).__call__(request)
        except HTTPException:
            credentials = None

        if credentials:
            if not credentials.scheme == "Bearer":
                save_to_logging_context({"scheme": credentials.scheme})
                logger.info(msg="Failed to authenticate user; Invalid authentication scheme.", extra=logging_context())
                raise HTTPException(status_code=403, detail="Invalid authentication scheme.")

            token_payload = self.verify_jwt(credentials.credentials)

            if not token_payload:
                logger.info(msg="Failed to authenticate user; Invalid or expired token.", extra=logging_context())
                raise HTTPException(status_code=403, detail="Invalid token or expired token.")

            logger.debug(msg="Successfully acquired JWT.", extra=logging_context())
            return token_payload

        else:
            logger.debug(msg="Failed to authenticate user; No credentials were provided.", extra=logging_context())
            return None

    @staticmethod
    def verify_jwt(token: str) -> dict:
        return decode_jwt(token)


####################################################################################################
# API key authentication
####################################################################################################

access_token_query = APIKeyQuery(name=ACCESS_TOKEN_NAME, auto_error=False)
access_token_header = APIKeyHeader(name=ACCESS_TOKEN_NAME, auto_error=False)
access_token_cookie = APIKeyCookie(name=ACCESS_TOKEN_NAME, auto_error=False)


async def get_access_token(
    # access_token_query: str = Security(access_token_query),
    access_token_header: Optional[str] = Security(access_token_header),
    access_token_cookie: Optional[str] = Security(access_token_cookie),
) -> Optional[str]:
    return access_token_header or access_token_cookie


async def get_current_user_data_from_api_key(
    db: Session = Depends(deps.get_db), access_token: str = Depends(get_access_token)
) -> Optional[UserData]:
    user = None
    roles: list[UserRole] = []

    if access_token is not None:
        access_key = db.query(AccessKey).filter(AccessKey.key_id == access_token).one_or_none()
        if access_key:
            user = access_key.user
            if access_key.role is not None:
                roles = [access_key.role] if access_key.role is not None else []

            save_to_logging_context(
                {
                    "user": user.id,
                    "active_roles": [role.name for role in roles],
                    "available_roles": [role.name for role in user.roles],
                }
            )

    if user is not None:
        logger.debug(msg="Successfully authenticated API key for user.", extra=logging_context())
        return UserData(user, roles)

    logger.debug(msg="Failed to authenticate user via API key; No key provided.", extra=logging_context())
    return None


####################################################################################################
# Main authentication methods
####################################################################################################


async def get_current_user(
    api_key_user_data: Optional[UserData] = Depends(get_current_user_data_from_api_key),
    token_payload: Optional[dict] = Depends(JWTBearer()),
    db: Session = Depends(deps.get_db),
    # Custom header for the role the authenticated user would like to assume.
    # Namespaced with x_ to indicate this is a custom application header.
    x_active_roles: Optional[str] = Header(default=None),
) -> Optional[UserData]:
    save_to_logging_context({"requested_roles": x_active_roles})

    if api_key_user_data is not None:
        save_to_logging_context({"auth_method": AuthenticationMethod.api_key, "user_authenticated": True})
        logger.info(msg="Successfully authenticated user via API key.", extra=logging_context())
        return api_key_user_data

    if token_payload is None:
        save_to_logging_context({"auth_method": None, "user_authenticated": False})
        logger.info(
            msg="Failed to authenticate user; Could not acquire credentials via API key or JWT.",
            extra=logging_context(),
        )
        return None

    username: Optional[str] = token_payload.get("sub")
    if username is None:
        save_to_logging_context({"auth_method": AuthenticationMethod.jwt, "user_authenticated": False})
        logger.info(msg="Failed to authenticate user; Username not present in token payload.", extra=logging_context())
        return None

    user = db.query(User).filter(User.username == username).one_or_none()

    # If there was a token payload, auth method must be JWT.
    save_to_logging_context({"auth_method": AuthenticationMethod.jwt})

    # A new user has just connected an ORCID iD. Create the user account.
    if user is None:
        # A new user has just connected an ORCID iD. Fetch their email address if it's visible, and create the
        # user account.
        email = fetch_orcid_user_email(username)
        user = User(
            username=username,
            is_active=True,
            # TODO When we decouple from the old database, change first_name and last_name to be nullable, and
            # stop filling them with empty strings.
            first_name=token_payload["given_name"] if "given_name" in token_payload else "",
            last_name=token_payload["family_name"] if "family_name" in token_payload else "",
            date_joined=datetime.now(),
            email=email,
            is_first_login=True,
        )
        save_to_logging_context(
            {
                "user": user.id,
                "first_login": user.is_first_login,
                "user_authenticated": True,
            }
        )

        logger.debug(msg="Created new user.", extra=logging_context())

    elif not user.is_active:
        save_to_logging_context({"user": user.id, "user_authenticated": True})
        logger.info(msg="Failed to authenticate user; User is inactive.", extra=logging_context())
        return None
    else:
        user.last_login = datetime.now()
        user.is_first_login = False
        save_to_logging_context(
            {
                "user": user.id,
                "first_login": user.is_first_login,
                "user_authenticated": True,
            }
        )

    db.add(user)
    db.commit()
    db.refresh(user)
    logger.info(msg="Successfully authenticated user via JWT.", extra=logging_context())

    # When no roles are requested, the user may act as any assigned role.
    save_to_logging_context({"available_roles": [role.name for role in user.roles]})
    if x_active_roles is None:
        save_to_logging_context({"active_roles": [role.name for role in user.roles]})
        logger.info(msg="Successfully assigned user roles to authenticated user.", extra=logging_context())
        return UserData(user, user.roles)

    # FastAPI has poor support for headers of type list (really, they are just comma separated strings).
    # Parse out any requested roles manually.
    requested_roles = x_active_roles.split(",")

    active_roles: list[UserRole] = []
    for requested_role in requested_roles:
        # Disregard any requested roles if they do not correspond to one of our UserRole enumerations.
        #
        # NOTE that our permissions structure ensures every authenticated user has a minimum set of available actions.
        # Collectively, these are known to the client as the 'ordinary user' role. Because these permissions are supplied
        # by default, we do not need add the 'ordinary user' role to the list of active roles.
        if requested_role not in UserRole._member_names_:
            logger.debug(msg=f"Ignoring unknown requested role {requested_role}.", extra=logging_context())
            continue

        enumerated_role = UserRole[requested_role]
        if enumerated_role not in user.roles:
            logger.warning(msg="User requested role to which they do not belong.", extra=logging_context())
            raise HTTPException(
                status_code=403,
                detail="This user is not a member of the requested acting role.",
            )
        else:
            active_roles.append(enumerated_role)

    save_to_logging_context({"active_roles": [role.name for role in active_roles]})
    logger.info(msg="Successfully assigned user roles to authenticated user.", extra=logging_context())
    return UserData(user, active_roles)



from datetime import date
from typing import Sequence

from pydantic.types import Optional

from mavedb.view_models import record_type_validator, set_record_type
from mavedb.view_models.base.base import BaseModel
from mavedb.view_models.contributor import Contributor
from mavedb.view_models.experiment import Experiment, ExperimentPublicDump, SavedExperiment
from mavedb.view_models.user import SavedUser, User


class ExperimentSetBase(BaseModel):
    urn: str
    published_date: Optional[date]


class ExperimentSetCreate(ExperimentSetBase):
    pass


class ExperimentSetUpdate(ExperimentSetBase):
    pass


# Properties shared by models stored in DB
class SavedExperimentSet(ExperimentSetBase):
    id: int
    record_type: str = None  # type: ignore
    experiments: Sequence[SavedExperiment]
    created_by: Optional[SavedUser]
    modified_by: Optional[SavedUser]
    creation_date: date
    modification_date: date
    contributors: list[Contributor]

    _record_type_factory = record_type_validator()(set_record_type)

    class Config:
        orm_mode = True


# Properties to return to non-admin clients
class ExperimentSet(SavedExperimentSet):
    created_by: Optional[User]
    modified_by: Optional[User]
    experiments: Sequence[Experiment]


# Properties to return to admin clients
class AdminExperimentSet(SavedExperimentSet):
    private: bool
    approved: bool
    processing_state: Optional[str]
    created_by: Optional[User]
    modified_by: Optional[User]
    experiments: Sequence[Experiment]


# Properties to include in a dump of all published data.
class ExperimentSetPublicDump(SavedExperimentSet):
    experiments: Sequence[ExperimentPublicDump]
    created_by: Optional[User]
    modified_by: Optional[User]



from sqlalchemy import Column, ForeignKey
from sqlalchemy.schema import Table

from mavedb.db.base import Base

collection_experiments_association_table = Table(
    "collection_experiments",
    Base.metadata,
    Column("collection_id", ForeignKey("collections.id"), primary_key=True),
    Column("experiment_id", ForeignKey("experiments.id"), primary_key=True),
)

collection_score_sets_association_table = Table(
    "collection_score_sets",
    Base.metadata,
    Column("collection_id", ForeignKey("collections.id"), primary_key=True),
    Column("score_set_id", ForeignKey("scoresets.id"), primary_key=True),
)



import re

MAVEDB_EXPERIMENT_SET_URN_DIGITS = 8
MAVEDB_URN_MAX_LENGTH = 64
MAVEDB_URN_NAMESPACE = "mavedb"

# Temp URN
MAVEDB_TMP_URN_PATTERN = r"tmp:[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}"
MAVEDB_TMP_URN_RE = re.compile(MAVEDB_TMP_URN_PATTERN)

# Old temp URN
MAVEDB_OLD_TMP_URN_PATTERN = r"^tmp:[A-Za-z0-9]{16}$"
MAVEDB_OLD_TMP_URN_RE = re.compile(MAVEDB_OLD_TMP_URN_PATTERN)

# Experiment set URN
MAVEDB_EXPERIMENT_SET_URN_PATTERN = rf"urn:{MAVEDB_URN_NAMESPACE}:\d{{{MAVEDB_EXPERIMENT_SET_URN_DIGITS}}}"
MAVEDB_EXPERIMENT_SET_URN_RE = re.compile(MAVEDB_EXPERIMENT_SET_URN_PATTERN)

# Experiment URN
MAVEDB_EXPERIMENT_URN_PATTERN = rf"{MAVEDB_EXPERIMENT_SET_URN_PATTERN}-([a-z]+|0)"
MAVEDB_EXPERIMENT_URN_RE = re.compile(MAVEDB_EXPERIMENT_URN_PATTERN)

# Score set URN
MAVEDB_SCORE_SET_URN_PATTERN = rf"{MAVEDB_EXPERIMENT_URN_PATTERN}-[1-9]\d*"
MAVEDB_SCORE_SET_URN_RE = re.compile(MAVEDB_SCORE_SET_URN_PATTERN)

# Variant URN
MAVEDB_VARIANT_URN_PATTERN = rf"{MAVEDB_SCORE_SET_URN_PATTERN}#[1-9]\d*"
MAVEDB_VARIANT_URN_RE = re.compile(MAVEDB_VARIANT_URN_PATTERN)

# Collection URN
MAVEDB_COLLECTION_URN_PATTERN = r"urn:mavedb:collection-[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}"
MAVEDB_COLLECTION_URN_RE = re.compile(MAVEDB_COLLECTION_URN_PATTERN)

# Any URN
MAVEDB_ANY_URN_PATTERN = "|".join(
    [
        rf"({p})"
        for p in (
            MAVEDB_TMP_URN_PATTERN,
            MAVEDB_EXPERIMENT_SET_URN_PATTERN,
            MAVEDB_EXPERIMENT_URN_PATTERN,
            MAVEDB_SCORE_SET_URN_PATTERN,
            MAVEDB_VARIANT_URN_PATTERN,
        )
    ]
)
MAVEDB_ANY_URN_RE = re.compile(MAVEDB_ANY_URN_PATTERN)



# See https://pydantic-docs.helpmanual.io/usage/postponed_annotations/#self-referencing-models
from __future__ import annotations

from typing import Optional

from mavedb.lib.validation import keywords
from mavedb.view_models import record_type_validator, set_record_type
from mavedb.view_models.base.base import BaseModel, validator


class KeywordBase(BaseModel):
    """Base class for keyword view models.
    Keywords may have key but no value if users don't choose anything from dropdown menu.
    TODO: Should modify it when we confirm the final controlled keyword list.
    """

    key: str
    value: Optional[str]
    vocabulary: Optional[str]
    special: Optional[bool]
    description: Optional[str]

    @validator("key")
    def validate_key(cls, v):
        keywords.validate_keyword(v)
        return v

    # validator("value") blocks creating a new experiment without controlled keywords so comment it first.
    # @validator("value")
    # def validate_value(cls, v):
    #     keywords.validate_keyword(v)
    #     return v


class KeywordCreate(KeywordBase):
    """View model for creating a new keyword."""

    pass


class KeywordUpdate(KeywordBase):
    """View model for updating a keyword."""

    pass


class SavedKeyword(KeywordBase):
    """Base class for keyword view models representing saved records."""

    record_type: str = None  # type: ignore

    _record_type_factory = record_type_validator()(set_record_type)

    class Config:
        orm_mode = True
        arbitrary_types_allowed = True


class Keyword(SavedKeyword):
    """Keyword view model for non-admin clients."""

    pass


class AdminKeyword(SavedKeyword):
    """Keyword view model containing properties to return to admin clients."""

    id: int



# Prevent circular imports
from typing import TYPE_CHECKING

from sqlalchemy import Boolean, Column, ForeignKey, Integer
from sqlalchemy.orm import Mapped, relationship

from mavedb.db.base import Base

if TYPE_CHECKING:
    from mavedb.models.experiment import Experiment
    from mavedb.models.publication_identifier import PublicationIdentifier


class ExperimentPublicationIdentifierAssociation(Base):
    __tablename__ = "experiment_publication_identifiers"

    experiment_id = Column(Integer, ForeignKey("experiments.id"), primary_key=True)
    publication_identifier_id = Column(Integer, ForeignKey("publication_identifiers.id"), primary_key=True)
    primary = Column(Boolean, nullable=True, default=False)

    experiment: Mapped["Experiment"] = relationship("Experiment", back_populates="publication_identifier_associations")
    publication: Mapped["PublicationIdentifier"] = relationship("PublicationIdentifier")
